/* -*- Mode: C++ -*- */

/*****************************************************************************
 * $RCSfile: BTeamAction.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *		       Jens Meyer, Robert Adolph, Volker Weinert
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/04/15 03:24:19 $
 * $Revision: 1.123 $
 *
 * TODO: move verbessern
 *
 *****************************************************************************/

#include "BTeamAction.hh"
#include "PFieldLogger.hh"
#include "Log.hh"
#include "kick.h"
#include "dribble.h"
#include "mybehave.h"
#include "client.h"
#include "PotentialField.hh"
#include "initpotentialloader.h"
#include "world.hh"
#include "CommunicationManager.hh"
#include <list>
#include "AttackBehaviour.hh"
#include "ScoringBehaviour.hh"
#include "DefaultBehaviour.hh"
#include "OffensiveMoveBehaviour.hh"


PlayerInterceptionInfo::PlayerInterceptionInfo() : distance(0.0),side(cmu::Mem->MySide),isUnableToIntercept(true) {};


BTeamAction::BTeamAction() : currentAction( 0 ), wait_for_kick(0), wait_for_their_kick(0) {
  world = new World();
  potentalField = new PotentialField(world);
  communication = new CommunicationManager();
  ipl = new InitPotentialLoader();

  attackBehaviour = new AttackBehaviour();
  scoringBehaviour = new ScoringBehaviour();
  defaultBehaviour = new DefaultBehaviour( &timing );
  offensiveMoveBehaviour = new OffensiveMoveBehaviour();

}

void BTeamAction::initialize()
{

  if ( world -> myNumber == 0 ) world->myNumber = cmu::Mem->MyNumber;

  if ( world->myNumber == 1 ) // falls ich Torwart bin will ich den Listen Modus verwenden
    ((DefaultBehaviour*)defaultBehaviour)->setListMode( true );

  setInitialValues(world->myNumber);
  // wenns was zu initialisieren gibt (vor dem Spielstart)
  // dann kann man das hier machen
  Log::warn.disable();
  Log::info.disable();
  Log::debug.disable();

  PFieldLogger::getInstance().setEnabled(false);
}

void BTeamAction::doSay(char *message)	 // Say Message
{
  cmu::Mem->SayAction.type = cmu::CMD_say;
  cmu::Mem->SayAction.time = cmu::Mem->CurrentTime;
  sprintf(cmu::Mem->SayAction.command, "(say %s)",message);
}

// To use an online coach may be very useful!
void BTeamAction::parse_received_message(int direction, char *message)
{
  communication->parse(message);
}

/*****************************************************************************/
/*********************	Normal Player Central Module  ************************/
/*****************************************************************************/

inline float BTeamAction::calculateAngle( float x1, float y1, float x2, float y2 ) {
  /* eingemachtes
   *
   * In welchem Winkel stehen die beiden Punkte zueinander ?
   */

  float dy = y2-y1;
  float dx = x2-x1;
  return atanh( dy/dx );

}


inline float BTeamAction::distance(cmu::Vector a,cmu::Vector b)
{
  return sqrt(((b.x-a.x)*(b.x-a.x))+((b.y-a.y)*(b.y-a.y)));
}




void BTeamAction::setInitialValues(cmu::Unum aNumber, bool reset) {

  static bool initializationDone = false;

  if (initializationDone && !reset) return;
  initializationDone = true;

  // Log::info << "setInitialValues" << endl;
  ipl = new InitPotentialLoader();
  char* filename;

  switch (aNumber) {
  case 1:
    filename = "Player1.txt";break;
  case 2:
    filename = "Player2.txt";break;
  case 3:
    filename = "Player3.txt";break;
  case 4:
    filename = "Player4.txt";break;
  case 5:
    filename = "Player5.txt";break;
  case 6:
    filename = "Player6.txt";break;
  case 7:
    filename = "Player7.txt";break;
  case 8:
    filename = "Player8.txt";break;
  case 9:
    filename = "Player9.txt";break;
  case 10:
    filename = "Player10.txt";break;
  case 11:
    filename = "Player11.txt";break;
  default : filename = "Player1.txt";break;

  }

  ipl->readPotentialFile(filename,potentialField());
}


void BTeamAction::normalPlayer() {

  /* Ausgaben: ich selbst und der Ball, eine wundersame Beziehung

  Log::debug << " " << endl << "====== cycle start/end ======" << endl;


  Log::debug << "[ich selbst] x:" << cmu::Mem->MyX() << " y:" << cmu::Mem->MyY() << endl;
  cmu::BallObject * bo = cmu::Mem->GetBall();
  if ( bo != NULL ) {

    cmu::Vector bp = bo->get_abs_pos();
    if ( bo->pos_valid() ) {
      Log::debug << "[der ball] x:" << bp.x << " y:" << bp.y << " ball conf:" << cmu::Mem->GetBall()->pos_valid() << " (ok)" << endl;
    } else {
      Log::debug << "[der ball] x:" << bp.x << " y:" << bp.y << " ball conf:" << cmu::Mem->GetBall()->pos_valid() << endl;
    }
  }
  */

  // ===========================
  // initialize game

  static bool players_moved = false;
  if (cmu::Mem->PlayMode == cmu::PM_Before_Kick_Off ||
      cmu::Mem->PlayMode == cmu::PM_Half_Time ||
      cmu::Mem->PlayMode == cmu::PM_Time_Up ||
      cmu::Mem->PlayMode == cmu::PM_Extended_Time) {

    if (players_moved) return;
    else {
      players_moved = true;
      cmu::move(PotentialField::translateToGlobalX(potentialField()->getSelfGrundX()),
		PotentialField::translateToGlobalY(potentialField()->getSelfGrundY()));
      return;
    }
  }

  if (cmu::Mem->BallSpeed()>maxBallSpeed) maxBallSpeed = cmu::Mem->BallSpeed();

  players_moved = false;
  PFieldLogger::getInstance().flush(false);

  // init currentAction with 0 every cycle.
  // if some action produces a currentAction, it will have a certain value at the end of
  // the cycle. if not, it will have 0 as value.
  currentAction = 0;

  RefreshPositioningMatrix();

  Log::debug << "scoring?" << endl;
  if (!scoringBehaviour->behave()) {

    Log::debug << "no, offensive move?" << endl;
    if (!offensiveMoveBehaviour->behave()) {

      Log::debug << "no, attack?" << endl;
      if (!attackBehaviour->behave()) {
	
	Log::debug << "no, reflex?" << endl;
	if (!reflex()) {
	
	  Log::debug << "no, potential field" << endl;
	  if ( !defaultBehaviour->behave() )

	    Log::warn << "NO DEFAULT BEHAVIOUR! => something is seriously wrong here" << endl;
	  
	}
      }
    }
  }

  Log::debug << "done acting" << endl;

  communication->broadcast();

}

bool BTeamAction::isMyStandardSituation() {

  // copied from world.cc, because we need it here and don't want
  // world.update() to run

  switch (cmu::Mem->PlayMode)
    {
    case cmu::PM_My_Kick_Off:
      return true;
    case cmu::PM_My_Kick_In:
      return true;
    case cmu::PM_My_Free_Kick:
      return true;
    case cmu::PM_My_Corner_Kick:
      return true;
    case cmu::PM_My_Goal_Kick:
      return true;
    case cmu::PM_My_Goalie_Free_Kick:
      return true;
    case cmu::PM_My_Offside_Kick:
      return true;
    default:
      return false;
    }
  return false;
}

bool BTeamAction::isTheirStandardSituation() {

  // copied from world.cc, because we need it here and don't want
  // world.update() to run

  switch (cmu::Mem->PlayMode)
    {
    case cmu::PM_Their_Kick_Off:
      return true;
    case cmu::PM_Their_Kick_In:
      return true;
    case cmu::PM_Their_Free_Kick:
      return true;
    case cmu::PM_Their_Corner_Kick:
      return true;
    case cmu::PM_Their_Goal_Kick:
      return true;
    case cmu::PM_Their_Goalie_Free_Kick:
      return true;
    case cmu::PM_Their_Offside_Kick:
      return true;
    default:
      return false;
    }
  return false;
}

/** diese Methode ermittelt, ob in der derzeitigen Situation eine
    simple Reaktion erforderlich ist, und fuehrt diese dann auch aus.
    gibt true zurueck, falls Reflex ausgefuehrt wurde.
    gibt sonst false zurueck */
bool BTeamAction::reflex() {
  // contains very simple reactive strategy overriding all
  // decisions. just feel it.

  /* warning: HACK
     in standard situations just stand still. otherwise the
     potential field would tell the player to run towards the ball keeping
     player, thus burning stamina without sense.

     there should be some other mechanism for this
  */

  // if (wait_for_their_kick >= 25)
  //  {

  //  Log::debug << "immer noch Gegner - Standardsituation - keine Aktion " << endl;
  //	return true;

  //} else

  static bool holdBall = false;
  static bool catchBall = false;
  static int lastScan;
  static int keptIP = 0;
  static float maxBV = 0;
  static float minBV = 0;

  if (!isMyStandardSituation()	  && (wait_for_kick>0)) 	 wait_for_kick	     = 0;
  if (!isTheirStandardSituation() && (wait_for_their_kick>0))	 wait_for_their_kick = 0;


  if (isMyStandardSituation() && (wait_for_kick<45))
    {
      if (cmu::Mem->MyNumber==1)
	{
	  wait_for_kick++;
	  return true;
	}

      wait_for_kick++;
      return false;
    }

  else if (isTheirStandardSituation() && (wait_for_their_kick<25))
    {
      Log :: debug << "Gegner - Standarsituation - ich warte auf Aktion" << endl;
      wait_for_their_kick++;
      return false;
    }

  else if (isTheirStandardSituation() && (wait_for_their_kick>=25))
    return true;


  /* ANMERKUNG:
     BallPathInterceptCycles funktioniert nicht so recht.
     die Werte scheinen sehr unsinnig.

     if ((cmu::Mem->NumPlayersCloserToBall() < 2) &&
	 (cmu::Mem->BallPathInterceptCyclesForBall() <
	  INTERCEPT_CYCLES_THRESHOLD) &&
	 (cmu::Mem->BallPathInterceptCanIGetThere(80)))
     {
  */

  if (!catchBall && cmu::Mem->BallCatchable() && ( cmu::Mem->MyNumber == 1)) {
      cmu::goalie_catch(cmu::Mem->BallAngleFromBody());
      catchBall = true;
      
	return true;
  } else if ( !cmu::Mem->BallCatchable() && ( cmu::Mem->MyNumber == 1) ) {

      catchBall = false;

  }
  
  cmu::Unum fastestTeammate = cmu::Mem->FastestTeammateToBall();
  cmu::Unum fastestOpponent = cmu::Mem->FastestOpponentToBall();

  int teammateCycles = (fastestTeammate == Unum_Unknown)?
    654565654:
      cmu::Mem->TeammateInterceptionNumberCycles(fastestTeammate);

  int opponentCycles = (fastestOpponent == Unum_Unknown)?
    654565654:
      cmu::Mem->OpponentInterceptionNumberCycles(fastestOpponent);

  int myCycles = cmu::Mem->MyInterceptionNumberCycles();

  // Korrektur von dummen Werten aus dem CMU Part mit anderen dummen Werten
  if (teammateCycles < 0) teammateCycles = 654565654;
  if (opponentCycles < 0) opponentCycles = 654565654;
  if (myCycles	     < 0) myCycles	 = 654565654;


  /*	Ausgaben fuer die Analyse *******


  if (fastestTeammate == Unum_Unknown)
    Log :: debug << "fastest teammate to ball unknown" << endl;
  else
    Log :: debug << "Fastest Teammate to Ball : " << fastestTeammate << " in " << teammateCycles << " Zyklen " << endl;

  if (fastestOpponent == Unum_Unknown)
    Log :: debug << "fastest opponent to ball unknown" << endl;
  else
    Log :: debug << "Fastest Opponent to Ball : " << fastestOpponent << " in " << opponentCycles << " Zyklen " << endl;

  int i=0;
  cmu::PlayerObject * tempPlayer = NULL;

  while (i < cmu::Mem->NumTeammates()) {
    if (i < cmu::Mem->NumTeammates()) {
      tempPlayer = cmu::Mem->Teammates()[i];

      if ((tempPlayer != NULL) &&
	  (tempPlayer->pos_valid()) &&
	  (cmu::Mem->MyNumber != tempPlayer->unum)) {
	Log :: debug << "Teammate " << tempPlayer->unum << ": pos=" << tempPlayer->get_abs_pos() << ", vel=" << tempPlayer->get_abs_vel() << endl;
      }
    }
    i++;
  }
  i=0;
  while (i < cmu::Mem->NumOpponents()) {
    if (i < cmu::Mem->NumOpponents()) {
      tempPlayer = cmu::Mem->Opponents()[i];

      if ((tempPlayer != NULL) &&
	  (tempPlayer->pos_valid())) {
	Log :: debug << "Opponent " << tempPlayer->unum << ": pos=" << tempPlayer->get_abs_pos() << ", vel=" << tempPlayer->get_abs_vel() << endl;
      }
    }
    i++;
  }

  Log::debug << "reflex(): ball confidence = " << cmu::Mem->GetBall()->pos_valid() << endl;

  // Ende Ausgaben ********* */

  if ((cmu::Mem->BallPositionValid() && cmu::Mem->MyConf())) {
    // Log :: debug << "MyConf == true" << endl;
    // dieses ist ein workaround wegen eines anscheinenden Bugs im CMU
    // Weltmodell. BallPathInterceptCyclesForBall() gibt mitunter den maximalen
    // int Wert zurueck, weshalb in der Folge BallPathInterceptCanIGetThere(80)
    // den Agenten zum Absturz bringt.


    // if (cmu::Mem->BallPathInterceptCyclesForBall() > 100) return false;


    cmu::Vector alterInterceptionPoint = interceptionPoint;

    interceptionPoint = cmu::Mem->MyInterceptionPoint();
    if (interceptionPoint==cmu::Vector(0,0)) {

      return false;
    }

    float distIP = distance(interceptionPoint,alterInterceptionPoint);
    float myDist = distance(cmu::Vector(cmu::Mem->MyX(),cmu::Mem->MyY()),interceptionPoint);
    float teammateDist = (fastestTeammate == Unum_Unknown)?654565654.0:
      cmu::Mem->distance(cmu::Mem->TeammateAbsolutePosition( fastestTeammate ),interceptionPoint);

    float ballVel = 0.0;
    if (cmu::Mem->BallVelocityValid()) ballVel = cmu::Mem->BallSpeed();

    if (ballVel > maxBV) maxBV = ballVel;
    if (ballVel < minBV) minBV = ballVel;


    /* Verwende alten InterceptionPoint, wenn
     * - der Abstand des alten zum neuen InterceptionPoint gering ist
     * - mein Abstand zum InterceptionPoint gross ist
     */
    if (((distIP<4.0) && (myDist > 5.0)) || ((distIP<2.0) && (myDist > 2.0)))
      {
	keptIP++;
	interceptionPoint = alterInterceptionPoint;
      } else keptIP = 0;

      // Log ::debug << "mein interception-point " << interceptionPoint << " in " << myCycles << " Zyklen" << endl;

    if ((myCycles <= teammateCycles)||((myDist-teammateDist) < -2.0)) {

      cmu::InterceptRes res = cmu::Mem->MyInterceptionResult (100);

      // ok, wir sind anscheinend am naechsten zum Ball,
      // und der ist recht nah dran.
      // also gehn wir hin

      // if (myCycles > opponentCycles)
      //	interceptionPoint = cmu::Mem->OpponentInterceptionPoint(fastestOpponent);

      if (res == cmu::BI_CanChase) {
	holdBall = false;

	/*
	if (((cmu::Mem->MyNumber==4) || (cmu::Mem->MyNumber==5)) &&
	    (interceptionAussen(interceptionPoint)) &&
	    (myCycles > opponentCycles))
	  {
	    // Log :: debug << " Ich bin Innenverteidiger, der Ball ist aussen, deswegen gehe ich nicht hin " << endl;
	    return false;
	  }


	if ((myCycles > opponentCycles) &&
	    (behindBallAndOpponent()) &&
	    (interceptionAussen(interceptionPoint)))
	  {
	    // Log :: debug << " Ich bin Aussenverteidiger, der Gegenspieler ist an mir vorbei, deswegen ziehe ich nach innen " << endl;
	    if (interceptionPoint.y>0.0)
	      interceptionPoint.y = interceptionPoint.y - 10.0;
	    else
	      interceptionPoint.y = interceptionPoint.y + 10.0;
	  }

	*/
	
	// so. sind noch nicht nah genug dran, laufen hin.
	
	if(cmu::Mem->MyNumber==1 && (interceptionPoint.y >=16.0 || interceptionPoint.y <= -16.0))return false;
	cmu::go_to_point( interceptionPoint , 0.0,
			  100,
			  cmu::DT_none);

	  //Log ::debug << "reflex :: get ball - interception Point" << interceptionPoint << endl;

	
	return true;
	
      }

      else if (!holdBall && cmu::Mem->BallCatchable()) {
	cmu::goalie_catch(cmu::Mem->BallAngleFromBody());
	holdBall = true;
	
	return true;
	//
      } else if (!holdBall && (res == cmu::BI_ReadyToKick)&&(cmu::Mem->BallMoving()) && (!dribbeln)) {
	holdBall = true;
	
	cmu::stop_ball(); // maybe use stop_ball ?
	return true;
      } else {
	return false;
      }
    }

  } else {

    if (cmu::Mem->CurrentTime.t > (lastScan+20)) {
      lastScan = cmu::Mem->CurrentTime.t;
      cmu::scan_field_with_body();
    }

  }

  return false;
}



bool BTeamAction::interceptionAussen(cmu::Vector interceptionPoint)
{
  return ((interceptionPoint.x<0.0) & ((interceptionPoint.y>15.0)  || (interceptionPoint.y<-15.0)));
}





/** Gegner ist naeher am Ball, das wurde bereits geprueft
    nun ist zu pruefen, ob der Ball sich von uns weg bewegt */
bool BTeamAction::behindBallAndOpponent()
{
  float dist_now = distance(cmu::Mem->MyPos(),cmu::Mem->BallAbsolutePosition());
  float dist_future = distance(cmu::Mem->MyPos(),cmu::Mem->BallPredictedPosition(3));
  return (dist_future>dist_now);
}


void BTeamAction::doKickToPoint( const cmu::Vector & target, float targ_vel ) {

  // for the broadcast. it should know which action we are going to do
  static Action action;

  action = Action( target );
  action.type = ACTION_kick;
  setCurrentAction( &action );

  Log::debug << "doKick to point " << target << endl;
  kick_ball( target, cmu::KM_Quickly, targ_vel, cmu::TURN_NONE );

  //  timing.registerKick();
}

void BTeamAction::doKick(const Action &action) {
  Log::debug << "doKick to " << (cmu::Vector( action.x, action.y )) << endl;

  // for the broadcast. it should know which action we are going to do
  static Action actionCopy;

  actionCopy = action;

  actionCopy.type = ACTION_kick;
  setCurrentAction( &actionCopy );

  if (timing.mayKick() || (cmu::Mem->PlayMode != cmu::PM_Play_On)) {
    //hier sollten wir entscheiden, ob der Schuss ein Pass oder ein Torschuss
    //sein soll. Bei letzterem muessen wir so hart wie moeglich schiessen, bei
    //ersterem hingegen so, dass der Ball an den angezielten Punkt mit einer
    //Geschwindigkeit ankommt, dass der Mitspieler ihn gut annehmen kann.

    if((isMyStandardSituation()) || (action.strength==1))
      {
	// if (action.strength==1)
	
	//smart_kick_hard(cmu::Mem->AngleToFromBody( cmu::Vector(action.x,action.y)), cmu::KM_Moderate);	
	kick_ball(cmu::Vector(action.x,action.y),cmu::KM_Quickly);
      }
    else{
      smart_pass(cmu::Vector(action.x, action.y), cmu::KM_Quickly);
    }

    timing.registerKick();
    //	  Log::debug << "kicked successfully (2)" << endl;
  }
  if (dribbeln)
    dribbeln = false;

}

void BTeamAction::doDribble(const Action &action) {
  Log::debug << "doDribble to " << (cmu::Vector( action.x, action.y )) << endl;

  cmu::Vector target( action.x, action.y );

  float angle = (target - cmu::Mem->MyPos()).dir();


  //  float angle =calculateAngle(cmu::Mem->MyX(),cmu::Mem->MyY(),action.x,action.y);

    if (isnan(angle))
    {
      Log ::debug << "angle war Nan - wir setzen ihn auf 180" << endl;
      angle=180.0;

    }
    cmu::kick(15,angle);
    dribbeln = true;

}


void BTeamAction::doGetball(const Action &action) {
  Log::debug << "doGetball to " << (cmu::Vector( action.x, action.y )) << endl;
  /*
    Log::info<<"PO_MyInterc before" <<PO_MyInterc<<endl;
    PO_MyInterc=10;
    // we want to get the ball. The decision for this has been made earlier
    Log::info<<"PO_MyInterc after" <<PO_MyInterc<<endl;

    PO_MyInterc=10;
    //we want to get the ball. The decision for this has been made earlier
  */
  /* if (PO_MyInterc-2 <= PO_TeammateInterc && PO_MyInterc <= 30 || PO_MyInterc <= 10){
     // cmu::get_ball();    }

     cmu::get_ball();
     else{  */

  // cmu::face_only_neck_to_ball();

  // diese wuerde keinen Sinn machen, wuerden wir nicht davon ausgehen,
  // dass das Potentialfeld im naechsten Zyklus erneut aufgerufen
  cmu::go_to_point(cmu::Vector(action.x,action.y),0.5,
		   //0.5+fabs(cmu::Mem->BallX()-cmu::Mem->MyX())/10.0,
		   100,
		   cmu::DT_unless_with_ball);

  // }

  // Log ::debug << "action:  get ball " << endl;
  timing.registerGetBall();
  if (dribbeln)
    dribbeln = false;

  //  }
}

float  BTeamAction::PowerConserveStamina(void)
{
  if (cmu::Mem->MyInterceptionAble() &&
      (abs(PO_MyInterc-PO_OpponInterc)<3 || PO_MyInterc<4)) return 100.0;
  if(cmu::Mem->MyStamina()-150.0<cmu::Mem->SP_recover_dec_thr*cmu::Mem->SP_stamina_max) return 30.0*(cmu::Mem->MyRecovery());
  return 100.0;
}

void BTeamAction::RefreshPositioningMatrix()
{
  if (cmu::Mem->FastestOpponentToBall()!=0) {
    PO_OpponInterc = cmu::Mem->OpponentInterceptionNumberCycles(cmu::Mem->FastestOpponentToBall());}
  else { PO_OpponInterc = 150; }
  if (cmu::Mem->FastestTeammateToBall() != 0) {
    PO_TeammateInterc = cmu::Mem->TeammateInterceptionNumberCycles(cmu::Mem->FastestTeammateToBall());}
  else { PO_TeammateInterc = 200; }
  if (cmu::Mem->MyInterceptionAble()) {
    PO_MyInterc = cmu::Mem->MyInterceptionNumberCycles(); }
  else { PO_MyInterc = 250; }
}



void BTeamAction::doRun(const Action &action) {
  Log::debug << "Laufe zu : " << cmu::Vector( action.x, action.y ) << endl;

  if(cmu::Mem->DistanceTo(cmu::Vector( action.x, action.y ))<=1.5)
  {
    if(cmu::Mem->DistanceTo(cmu::Mem->BallAbsolutePosition())>15.0){
      cmu::face_neck_to_ball();//scan_field_w... wird danach aufgerufen
    }
    else{
      cmu::go_to_point(cmu::Vector(action.x,action.y),1.5+fabs(cmu::Mem->BallX()-cmu::Mem->MyX())/10.0,PowerConserveStamina(),cmu::DT_none);
      cmu::face_only_neck_to_ball();
    }

  } else {
    cmu::go_to_point(cmu::Vector(action.x,action.y),1.5+fabs(cmu::Mem->BallX()-cmu::Mem->MyX())/10.0, PowerConserveStamina(),cmu::DT_none);
    cmu::face_only_neck_to_ball();
  }
  if (dribbeln)
    dribbeln = false;
}

void BTeamAction::doMove(const Action &action) {
  Log::debug << "doMove to " << (cmu::Vector( action.x, action.y )) << endl;
  cmu::move(cmu::Vector(action.x,action.y));
  if (dribbeln)
    dribbeln = false;
}

// Konstruktor fuer Action aus BTeamTypes.hh
Action::Action(int action[]) : valid(true) {
  type = (ActionType)(action[2]+1);
  pf_x = action[0];
  pf_y = action[1];
  x = PotentialField::translateToGlobalX(pf_x);
  y = PotentialField::translateToGlobalY(pf_y);
  strength = action[3];
}
// Konstruktor fuer Action aus BTeamTypes.hh
Action::Action(const cmu::Vector & v) : valid(true) {
  type = ACTION_unknown;
  x = v.x;
  y = v.y;
  strength = 1;
}

/****************************************************************************************/
/*************************  Goalie Player Central Module  *******************************/
/****************************************************************************************/

// This is a very simple goalie. A decent goalie [Reis and lau 2001] must look always (using 45
// degree vision) to the ball and must not try risky interceptions. Backward dashing is needed!
// Try to see if the opponent has the ball controled and use this on the reasoning process!
void BTeamAction::goaliePlayer()
{


  if (cmu::Mem->PlayMode == cmu::PM_Before_Kick_Off ||
      cmu::Mem->PlayMode == cmu::PM_Half_Time ||
      cmu::Mem->PlayMode == cmu::PM_Time_Up ||
      cmu::Mem->PlayMode == cmu::PM_Extended_Time ||


      cmu::Mem->BallAbsolutePosition().x <=5.0)
    {
     normalPlayer();
    } else {
      cmu::face_neck_to_ball();
    }



     // hier kommt das Verhalten des Goalies hin
     /*
     Aufgabe:
     Man formuliere doch mal einen Entscheidungsbaum in deutscher Sprache.

      ...

     */

}


/*
 * $Log: BTeamAction.cc,v $
 * Revision 1.123  2002/04/15 03:24:19  fastjack
 * BallMessage Klasse fr das Parsen/Generieren von BallInfos gebaut. CommunicationManager aufgerumt.
 *
 * Revision 1.122  2002/04/13 07:59:16	jens
 * German Open 2002 Edition
 *
 * Revision 1.121  2002/04/10 13:24:53	seekamp
 * Dribbling/Schuss verbessert.
 * Steilpaesse werden weniger hart gespielt.
 * Wenn aussen wird Richtung, Eckfahne gedribbelt
 *
 * Revision 1.120  2002/04/09 19:23:13	fastjack
 * Verbesserungen der Dribble und Torschuss Methoden. Bug fixes.
 *
 * Revision 1.119  2002/04/09 12:28:58	seekamp
 * Dribbling Richtung Torwart, wenn ClosestPlayerToTheirGoal
 *
 * Revision 1.118  2002/04/09 00:27:44	fastjack
 * kick improved
 *
 * Revision 1.117  2002/04/08 13:32:08	fastjack
 * listMode jetzt als Property von DefaultBehaviour. per default auf false, Torwart setzt auf true.
 *
 * Revision 1.116  2002/04/08 13:28:05	seekamp
 * Aktionsausfuehrung fuer Manndecker wieder Staminschonender implementiert
 *
 * Revision 1.115  2002/04/05 16:18:45	fastjack
 * OffensiveMoveBehaviour implemented. Under certain conditions ( which still may/should be improved ) our agent should dribble forward instead of shooting.
 *
 * Revision 1.114  2002/04/04 21:20:37	fastjack
 * bug fix
 *
 * Revision 1.113  2002/04/04 21:15:43	fastjack
 * removed PotentialField code from BTeamAction, all is in DefaultBehaviour now. I hope it works.
 *
 * Revision 1.112  2002/04/04 15:54:07	seekamp
 * kleinere Veraenderungen am Attack und Shot Behaviour zudem Libero wieder eingefuehrt und Aussenverteidiger bei gegnerischen Angriffen ueber die gegenueberliegende Seite Richtung Mitte gezogen.
 *
 * Revision 1.111  2002/04/02 17:06:51	fastjack
 * Scoring Behaviour ungetesterweise eingebaut.
 *
 * Revision 1.110  2002/04/02 15:58:10	fastjack
 * Various changes. Torwart Abschlag Bugfix. kleinere unbedeutendere Korrekturen.
 *
 * Revision 1.109  2002/04/02 15:52:49	seekamp
 * Positionsfindung der Abwehrspieler besser auf Abseitsfalle und Viererkette eingestellt.
 *
 * Revision 1.108  2002/04/01 15:57:11	fastjack
 * world.update wird nicht mehr im PotentialField aufgerufen. Grund: mehr Flexibilitaet. AttackBehaviour und BTeamAction von World entkoppelt.
 *
 * Revision 1.107  2002/04/01 14:56:48	seekamp
 * Verhalten bei Standardsituationen (eigene und gegnerische) verbessert.
 *
 * Revision 1.106  2002/03/28 16:19:45	seekamp
 * Anderes Verhalten vor Standardsituationen: es wird zunaechst 10 Zyklen gewartet, dann laeuft der ballnaechste Spieler zum Ball und schiesst direkt.
 *
 * Revision 1.105  2002/03/28 15:49:46	fastjack
 * Ball rumgeschreie vom FCP Agent eingebaut. Ausgaben wieder ausgebaut
 *
 * Revision 1.104  2002/03/28 12:52:11	fastjack
 * Absturzursache eingegrenzt - Player2Player Kommunikation entfernt als schnelle Loesung
 *
 * Revision 1.103  2002/03/26 17:56:22	fastjack
 * Performance Optimierungen.
 *
 * Revision 1.102  2002/03/26 17:32:11	seekamp
 * Es muessten jetzt Steilpaesse gespielt werden, konnte allerdings noch nicht getestet werden.
 *
 * Revision 1.101  2002/03/26 00:17:54	fastjack
 * test code optimized
 *
 * Revision 1.100  2002/03/25 17:32:21	seekamp
 * Testlaeufe
 *
 * Revision 1.99  2002/03/25 16:14:06  fastjack
 * Merged & added Behaviour classes for offensive behaviour
 *
 * Revision 1.98  2002/03/22 11:18:44  seekamp
 * neue Aufstellung
 *
 * Revision 1.97  2002/03/21 22:53:43  fastjack
 * bug fixes
 *
 * Revision 1.96  2002/03/21 22:32:08  fastjack
 * Methodennamen verbessert. Schnelle und ungenaue Methode zum Pruefen, ob ein Gegner einen Schuss abfangen kann, implementiert. Ungetestet.
 *
 * Revision 1.95  2002/03/21 18:25:50  fastjack
 * Wieder die Auswahl von mehreren Punkten eingefuegt. Methode zur optimalen Ermittlung der Ball Velocity entwickelt.
 *
 * Revision 1.94  2002/03/21 17:54:39  seekamp
 * Flanken sollen durch reflex-Methode von Aussenverteidigern abgefangen werden,
 *
 * Revision 1.93  2002/03/21 15:13:02  fastjack
 * reflex() aufgeraeumt. Ordnung muss sein :o)
 *
 * Revision 1.92  2002/03/21 15:02:27  seekamp
 * reflex() Methode optimiert.
 *
 * Revision 1.91  2002/03/18 15:53:42  seekamp
 * Abseitsfalle wiedereingefuehrt - Spielgeschehen nach vorne verlagert
 *
 * Revision 1.90  2002/03/15 14:43:26  fastjack
 * merged
 *
 * Revision 1.89  2002/03/15 14:38:36  seekamp
 * Absturzursachen beseitigt
 *
 * Revision 1.88  2002/03/15 13:56:01  seekamp
 * interception-Point veraendert
 *
 * Revision 1.87  2002/03/15 13:26:00  fastjack
 * Ausgaben entfernt / dem Gouraud Kerl mehr Genauigkeit verschafft und einen boesen BUG entfernt
 *
 * Revision 1.86  2002/03/14 17:43:15  seekamp
 * Fehlersuch im cone und bei der Aktionswahl
 *
 * Revision 1.85  2002/03/12 17:03:40  seekamp
 * Liste der Aktionen sortiert und Fehler im cone beseitigt
 *
 * Revision 1.84  2002/03/12 15:45:29  jens
 * Grundlage fuer das Lernen der Prioritaeten implementiert. Fehler dabei ausgemerzt und Algorithmus fuer das Lernen auf Papier konkretesiert.
 *
 * Revision 1.83  2002/03/11 15:36:14  fastjack
 * debug log disabled
 *
 * Revision 1.82  2002/03/11 15:16:28  fastjack
 * Wieder umgebaut nach Liste-erzeugen
 *
 * Revision 1.81  2002/03/11 13:50:34  fastjack
 * Alternative Zielpunkt-Pruefung eingebaut (alte Methode ist action1(...), neue ist action2(...)).
 *
 * Revision 1.80  2002/03/11 11:55:56  fastjack
 * bugfix: Spieler haben falsche Player??.txt eingelesen
 *
 * Revision 1.79  2002/03/08 16:24:31  jens
 * Arbeiten beendet, die uns Potentialfeldszenen im Client parsen lassen und uns dann vollstaendige Potentialfelder genriert. Potentialfelder die generiert wurden und welche, die wir als Szene genommen haben (also die Feldwerte weglassen) stimmen nun ueberein. Rundungsfehler sind beseitigt. Somit sollte die Grundlage fuer die automatische Prioritaetenanpassung gelegt sein.
 *
 * Revision 1.78  2002/03/08 12:58:00  fastjack
 * code wieder bereinigt
 *
 * Revision 1.77  2002/03/07 17:00:14  fastjack
 * Habe PotentialField so umgestellt, dass es wieder so funktioniert wie vor unserer Aenderung von gestern; Code in extra Methoden ausgelagert wegen uebersicht - und einen derzeit nicht aufgerufenen Glaettungsalgorithmus implementiert, muss noch getestet werden.
 *
 * Revision 1.76  2002/03/05 18:13:15  fastjack
 * Interception Methode noch etwas fertiger.
 *
 * Revision 1.75  2002/03/05 17:08:00  fastjack
 * Interception Methoden fast fertig
 *
 * Revision 1.74  2002/03/04 16:02:10  fastjack
 * Interception Points angefangen zu implementieren. ACHTUNG code nicht lauffaehig
 *
 * Revision 1.73  2002/03/04 14:39:45  seekamp
 * PotentialField gibt jetzt eine Liste von Aktionen zurueck und nicht nur einen Int-Wert
 *
 * Revision 1.72  2002/03/04 12:38:24  fastjack
 * Beispiel code in BTeamTypes.hh
 *
 * Revision 1.71  2002/03/01 12:56:41  seekamp
 * Abschlaege nicht mehr direkt zu den gegnerischen Stuermern
 *
 * Revision 1.70  2002/03/01 11:18:44  fastjack
 * GPL added
 *
 * Revision 1.69  2002/03/01 01:25:40  fastjack
 * start scripts und Makefile angepasst
 *
 * Revision 1.68  2002/02/28 16:59:54  seekamp
 * Absturzursache in einem kick-Befehl beseitigt
 *
 * Revision 1.67  2002/02/28 16:13:58  fastjack
 * bug fixed: Sporadische Abstuerze
 *
 * Revision 1.66  2002/02/28 11:56:48  fastjack
 * mh
 *
 * Revision 1.64  2002/02/26 19:43:27  seekamp
 * Interception-Point verbessert
 *
 * Revision 1.63  2002/02/26 17:59:29  jens
 * weiter am parsen eines eigenen PF gearbeitet. Einbinden in das Weltmodell funzt noch nicht richtig
 *
 * Revision 1.62  2002/02/25 18:06:38  seekamp
 * Interc.-Point - Verfolgung verbessert
 *
 * Revision 1.61  2002/02/25 12:15:32  fastjack
 * Abfangen von Durchgebrochenen Gegnern eingebaut.
 *
 * Revision 1.60  2002/02/21 17:24:57  seekamp
 * cone-Groesse veraendert
 *
 * Revision 1.59  2002/02/21 12:10:42  fastjack
 * Absturzproblem in CommunicationManager::send_game_status behoben
 *
 * Revision 1.58  2002/02/19 17:47:47  seekamp
 * +ausgaben
 *
 * Revision 1.57  2002/02/19 17:24:57  seekamp
 * +paar ausgaben
 *
 * Revision 1.56  2002/02/19 16:43:04  seekamp
 * blabla :-D
 *
 * Revision 1.55  2002/02/18 17:37:17  seekamp
 * Fehler im cone behoben
 *
 * Revision 1.54  2002/02/15 17:45:35  fastjack
 * PotentialField vorbereitet fuer Automatisierung durch zukuenftige Lernmethode, verschiedene Dinge aus World statt aus Mem geholt etc.
 *
 * Revision 1.53  2002/02/15 17:33:52  seekamp
 * neuen cone getestet - verbessertes Abwehrverhalten
 *
 * Revision 1.52  2002/02/15 13:51:09  fastjack
 * GouraudFiller repariert, fehlerhaften Algorithmus gefixt. Paar Kommentare entfernt.
 *
 * Revision 1.51  2002/02/14 19:39:53  fastjack
 * Gouraud Shader verbessert - sind noch Probleme drin leider.
 *
 * Revision 1.50  2002/02/14 19:05:41  seekamp
 * Abschlag variabel gestaltet. Abwehrstellungspiel veraendert. Schussart bei Standardsituationen ausgetauscht ( QuickestRelease)
 *
 * Revision 1.49  2002/02/11 17:04:15  seekamp
 * Fehlpassanalyse mit Wertanpassungen
 *
 * Revision 1.48  2002/02/10 10:58:18  fastjack
 * Habe den Code, der sich mit der Kommunikation befasst, in eine neue Klasse
 * CommunicationManager herausgezogen, da die BTeamAction sonst zu gross und
 * unbersichtlich wird.
 *
 * Revision 1.47  2002/02/07 20:07:50  fastjack
 * merged
 *
 * Revision 1.46  2002/02/07 19:41:33  jens
 * Kommunikation auf Mitspieler und im Falle eines Schusses auch den Zielpunkt erweitert.
 *
 * Revision 1.45  2002/02/05 19:14:01  fastjack
 * Versucht, den Interception-Reflex zu optimieren.
 *
 * Revision 1.44  2002/02/05 18:22:23  seekamp
 * ne menge sachen gemacht
 *
 * Revision 1.43  2002/02/05 18:19:39  jens
 * Fixed the Parse Sound. Now we can hear the Ball and the opponents broadcasted by other players.
 *
 * Revision 1.42  2002/02/05 13:41:07  fastjack
 * gemerged
 *
 * Revision 1.41  2002/02/05 12:27:40  jens
 * Weltmodell update
 *
 * Revision 1.40  2002/02/05 12:10:27  seekamp
 * Abseitslinie nach hinten verschoben. Abwehrspieler laufen sich bei Teamballbesitz nicht mehr frei, sondern halten ihre Position. Insgesamt defensivere Ausrichtung.
 *
 * Revision 1.39  2002/02/04 19:44:09  seekamp
 * kleine Verbesserungen am Torschuss und an der Abseitsfalle
 *
 * Revision 1.38  2002/01/31 18:22:26  fastjack
 * Versuch einer getBall Optimierung
 *
 * Revision 1.37  2002/01/30 13:05:30  seekamp
 * move-Befehl des Torwarts vorm Abschlag eingefuehrt
 *
 * Revision 1.36  2002/01/29 19:34:00  fastjack
 * Konflikte behoben.
 *
 * Revision 1.35  2002/01/29 19:18:51  seekamp
 * Verhalten fuer das Ausfuehren von Standardsituationen implementiert
 *
 * Revision 1.34  2002/01/29 13:55:49  fastjack
 * debugging still in progress
 *
 * Revision 1.33  2002/01/29 12:06:37  fastjack
 * debugging in progress
 *
 * Revision 1.32  2002/01/29 11:13:35  fastjack
 * Timing update
 *
 * Revision 1.31  2002/01/29 10:57:57  jens
 * kick verbessert und pausen der Spieler teilweise behoben (scan field with body eingesetzt)
 *
 * Revision 1.30  2002/01/28 17:22:03  seekamp
 * Dem Torwart Abschlaege und Abstoesse beigebracht
 *
 * Revision 1.29  2002/01/28 02:13:49  fastjack
 * Added an intermediary World object which should guess positions when it is done. Not ready yet, WARNING - code may or may not work.
 *
 * Revision 1.28  2002/01/25 18:54:38  seekamp
 * Staminaverwaltung implementiert
 *
 * Revision 1.26  2002/01/24 10:47:18  radolph
 * Uebermittlung der Ballposition falls der Spieler den Ball sieht. Sichergestellt, dass andere Spieler diese Information bekommen und gegebenenfalls verwenden.
 *
 * Revision 1.25  2002/01/22 19:05:10  seekamp
 * Torwart verbessert
 *
 * Revision 1.24  2002/01/22 18:04:00  jens
 * Some Staminaproblems solved
 *
 * Revision 1.23  2002/01/22 14:07:51  fastjack
 * Timing verbessert
 *
 * Revision 1.22  2002/01/22 12:06:00  fastjack
 * Timing tests
 *
 * Revision 1.21  2002/01/22 11:01:54  fastjack
 * BTeamAction refactoring
 *
 * Revision 1.20  2002/01/22 09:35:47  seekamp
 * Torwart und Feldgroesse
 *
 * Revision 1.19  2002/01/21 12:08:36  fastjack
 * PFieldLogger: das Format so geaendert, dass der PFViewer die PFelder auch parsen kann. Deshalb die UNUM erstmal wieder entfernt
 *
 * Revision 1.18  2002/01/18 20:11:20  fastjack
 * Das Logging von Potentialfeldern fertig gestellt. Das Logging ist auch
 * ein-/ausschaltbar ( PFieldLogger::getInstance().setEnabled(true) ).
 *
 * Revision 1.17  2002/01/18 19:49:28  seekamp
 * Prioritaeten werden durch Andys FileReader aus individuellen Spielertextdateien eingelesen.
 *
 * Revision 1.16  2002/01/17 19:11:40  fastjack
 * kleinere Aenderungen bzgl. pfield logging eingefuegt. dummy methode in PotentialField eingefuegt ...
 *
 * Revision 1.15  2002/01/17 18:15:21  seekamp
 * Verschiedene Methoden in Potentialfield und BTeamAction veraendert, mit dem
 * Ergebnis, dass der ballnaechste Spieler zum Ball laeuft und einen Pass spielt.
 * (mehr oder weniger)
 *
 * Revision 1.14  2002/01/15 21:50:28  seekamp
 * Fehler in der Bewertungsfunktion ausgemerzt.
 * Individuelles Spielverhalten getestet.
 *
 * Revision 1.13  2002/01/15 16:42:37  seekamp
 * Fehler in der Potentialfield.cc und in BTeamAction.cc beseitigt und Werte
 * veraendert in Poentialfield.hh
 *
 * Revision 1.12  2002/01/10 19:02:25  jens
 * tries to find further bugs. The actual bugs are: the Ballposition seems to be invalid sometimes. We dont know where the ballposition becomes set and why we have an invalid position although we have a valid position IN the Worldmodel of CMU. The second bug we tried to fix is an invalid position of player. It seems that only the last player has an invalid position.
 *
 * Revision 1.11  2002/01/07 18:50:05  jens
 * Further tryings to get rid of those Bugs. Found two sources, eliminated one
 * Actual Bug is located in the cenvesrion from RCC-Flags to CMU-Markers.
 *
 * Revision 1.10  2002/01/03 17:48:02  fastjack
 * einige abstuerze behoben.
 *
 * Revision 1.9  2002/01/03 01:04:36  fastjack
 * Parser verwendet fuer hear Messages wieder die originale CMU Methode. Ausserdem habe ich einige weitere Ausgaben in das PotentialField getan, um herauszufinden, wo er aussteigt. Dann habe ich das Timing so geHACKT, dass es immer true gibt, ausser es ist geblockt. Debugging necessary.
 *
 * Revision 1.8  2001/12/28 17:12:37  fastjack
 * joah
 *
 * Revision 1.7  2001/12/28 16:10:30  fastjack
 * verschiedenste bugfixes in allen bereichen
 *
 * Revision 1.6  2001/12/21 17:08:05  fastjack
 * Implemented rest of lowlevel functionality, if it really works cannot be tested, yet.
 *
 * Revision 1.5  2001/12/20 18:16:23  jens
 * eleminated different compile errors and started to implement simple actions
 *
 * Revision 1.4  2001/12/18 18:43:10  fastjack
 * Compiled nicht. Vorsicht, habe eine Menge auskommentiert. Timing z.B., und alles wo cmu:: drin vorkommt. Kruzifix!
 *
 * Revision 1.3  2001/12/14 20:41:39  fastjack
 * PotentialField neu geschrieben auf Basis von Matthias' Code
 *
 * Revision 1.2  2001/12/13 20:37:56  fastjack
 * update. k.a. wassas war :-)
 *
 * Revision 1.1  2001/12/13 01:42:10  fastjack
 * Ersatz fuer FCP Klasse
 *
 *
*/
