/* -*- Mode: C++ -*- */

/*****************************************************************************
 * $RCSfile: BTeamAction.hh,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *		       Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: jens $
 * $Date: 2002/04/13 07:59:16 $
 * $Revision: 1.46 $
 *****************************************************************************/

#ifndef _BTEAMACTION_HH_
#define _BTEAMACTION_HH_
#include "initpotentialloader.h"
#include "Timing.hh"
#include "BTeamTypes.hh"
#include "math.h"
#include "geometry.h"
#include <list>

namespace cmu
{
  typedef int Unum;
  class PlayerObject;
}

class PotentialField;
// class Timing;
class World;
class CommunicationManager;
class Behaviour;


class BTeamAction {

  private:
  static const float INTERCEPT_DISTANCE_THRESHOLD = 30.0;
  static const int INTERCEPT_CYCLES_THRESHOLD = 20;
  cmu::Vector interceptionPoint;
  bool dribbeln;
  int PO_OpponInterc;
  int PO_TeammateInterc;
  int PO_MyInterc;
  float maxBallSpeed;
  PotentialField * potentalField;
  Timing timing;
  InitPotentialLoader * ipl;
  Action * currentAction;
  int wait_for_kick;
  int wait_for_their_kick;

  // Behaviours
  Behaviour * attackBehaviour;
  Behaviour * scoringBehaviour;
  Behaviour * defaultBehaviour;
  Behaviour * offensiveMoveBehaviour;

  CommunicationManager * communication;
  World * world;

  /** True if game is on hold because of some standard situation for our team. */
  bool isMyStandardSituation();

  /** True if game is on hold because of some standard situation for the opponents team. */
  bool isTheirStandardSituation();

  public:
  BTeamAction();
  float PowerConserveStamina(void);
  void RefreshPositioningMatrix();
  inline float calculateAngle( float x1, float y1, float x2, float y2 );
  inline float distance (cmu::Vector a,cmu::Vector b);


  /**
  Triggers the behaviour for the player.
  */
  void normalPlayer();

  /**
  Triggers the behaviour for the goalie.
  */
  void goaliePlayer();

  /**
  Has to be invoked after start of the agent but before start of the game.
  */
  void initialize();


  bool interceptionAussen(cmu::Vector interceptionPoint);
  bool behindBallAndOpponent();

  bool reflex();

  void parse_received_message(int direction,char *message);
  void setInitialValues(cmu::Unum aNumber, bool reset = false);



  // =========== actions

  void doKickToPoint( const cmu::Vector & target, float targ_vel = 0.8 );

  void doKick(const Action &action);
  void doDribble(const Action &action);
  void doGetball(const Action &action);
  void doRun(const Action &action);
  void doMove(const Action &action);
  void doSay(char *message);


  // =========== getter and setter methods

  PotentialField * potentialField() { return potentalField; }

  Timing * getTiming() { return &timing; }

  World * getWorld() { return world; }

  Behaviour * getAttackBehaviour() { return attackBehaviour; }

  Behaviour * getScoringBehaviour() { return scoringBehaviour;	}

  Behaviour * getDefaultBehaviour() { return defaultBehaviour; }

  Behaviour * getOffensiveMoveBehaviour() { return offensiveMoveBehaviour; }

  Action * getCurrentAction() { return currentAction; }
  void setCurrentAction( Action * a ) { currentAction = a; }

  CommunicationManager * getCommunication() { return communication; }

};

#endif

/*
 * $Log: BTeamAction.hh,v $
 * Revision 1.46  2002/04/13 07:59:16  jens
 * German Open 2002 Edition
 *
 * Revision 1.45  2002/04/09 19:23:14  fastjack
 * Verbesserungen der Dribble und Torschuss Methoden. Bug fixes.
 *
 * Revision 1.44  2002/04/09 00:27:45  fastjack
 * kick improved
 *
 * Revision 1.43  2002/04/05 16:18:46  fastjack
 * OffensiveMoveBehaviour implemented. Under certain conditions ( which still may/should be improved ) our agent should dribble forward instead of shooting.
 *
 * Revision 1.42  2002/04/04 21:20:37  fastjack
 * bug fix
 *
 * Revision 1.41  2002/04/04 21:15:43  fastjack
 * removed PotentialField code from BTeamAction, all is in DefaultBehaviour now. I hope it works.
 *
 * Revision 1.40  2002/04/04 20:37:43  fastjack
 * bug fix
 *
 * Revision 1.39  2002/04/04 20:16:17  fastjack
 * moved PotentialField Behaviour in DefaultBehaviour class
 *
 * Revision 1.38  2002/04/02 17:06:51  fastjack
 * Scoring Behaviour ungetesterweise eingebaut.
 *
 * Revision 1.37  2002/04/01 15:57:12  fastjack
 * world.update wird nicht mehr im PotentialField aufgerufen. Grund: mehr Flexibilitaet. AttackBehaviour und BTeamAction von World entkoppelt.
 *
 * Revision 1.36  2002/04/01 14:56:48  seekamp
 * Verhalten bei Standardsituationen (eigene und gegnerische) verbessert.
 *
 * Revision 1.35  2002/03/28 16:19:45  seekamp
 * Anderes Verhalten vor Standardsituationen: es wird zunaechst 10 Zyklen gewartet, dann laeuft der ballnaechste Spieler zum Ball und schiesst direkt.
 *
 * Revision 1.34  2002/03/26 17:32:11  seekamp
 * Es muessten jetzt Steilpaesse gespielt werden, konnte allerdings noch nicht getestet werden.
 *
 * Revision 1.33  2002/03/21 22:53:43  fastjack
 * bug fixes
 *
 * Revision 1.32  2002/03/21 22:32:08  fastjack
 * Methodennamen verbessert. Schnelle und ungenaue Methode zum Pruefen, ob ein Gegner einen Schuss abfangen kann, implementiert. Ungetestet.
 *
 * Revision 1.31  2002/03/21 18:25:50  fastjack
 * Wieder die Auswahl von mehreren Punkten eingefuegt. Methode zur optimalen Ermittlung der Ball Velocity entwickelt.
 *
 * Revision 1.30  2002/03/21 17:54:39  seekamp
 * Flanken sollen durch reflex-Methode von Aussenverteidigern abgefangen werden,
 *
 * Revision 1.29  2002/03/15 13:26:00  fastjack
 * Ausgaben entfernt / dem Gouraud Kerl mehr Genauigkeit verschafft und einen boesen BUG entfernt
 *
 * Revision 1.28  2002/03/11 13:50:34  fastjack
 * Alternative Zielpunkt-Pruefung eingebaut (alte Methode ist action1(...), neue ist action2(...)).
 *
 * Revision 1.27  2002/03/05 17:08:00  fastjack
 * Interception Methoden fast fertig
 *
 * Revision 1.26  2002/03/04 16:02:13  fastjack
 * Interception Points angefangen zu implementieren. ACHTUNG code nicht lauffaehig
 *
 * Revision 1.25  2002/03/01 11:18:44  fastjack
 * GPL added
 *
 * Revision 1.24  2002/02/26 17:59:30  jens
 * weiter am parsen eines eigenen PF gearbeitet. Einbinden in das Weltmodell funzt noch nicht richtig
 *
 * Revision 1.23  2002/02/25 18:06:38  seekamp
 * Interc.-Point - Verfolgung verbessert
 *
 * Revision 1.22  2002/02/21 17:24:57  seekamp
 * cone-Groesse veraendert
 *
 * Revision 1.21  2002/02/19 16:43:04  seekamp
 * blabla :-D
 *
 * Revision 1.20  2002/02/10 10:58:18  fastjack
 * Habe den Code, der sich mit der Kommunikation befasst, in eine neue Klasse
 * CommunicationManager herausgezogen, da die BTeamAction sonst zu gross und
 * unbersichtlich wird.
 *
 * Revision 1.19  2002/02/07 19:41:34  jens
 * Kommunikation auf Mitspieler und im Falle eines Schusses auch den Zielpunkt erweitert.
 *
 * Revision 1.18  2002/02/05 19:14:01  fastjack
 * Versucht, den Interception-Reflex zu optimieren.
 *
 * Revision 1.17  2002/02/05 13:41:07  fastjack
 * gemerged
 *
 * Revision 1.16  2002/01/31 18:22:26  fastjack
 * Versuch einer getBall Optimierung
 *
 * Revision 1.15  2002/01/30 13:05:30  seekamp
 * move-Befehl des Torwarts vorm Abschlag eingefuehrt
 *
 * Revision 1.14  2002/01/28 02:13:49  fastjack
 * Added an intermediary World object which should guess positions when it is done. Not ready yet, WARNING - code may or may not work.
 *
 * Revision 1.13  2002/01/24 10:47:18  radolph
 * Uebermittlung der Ballposition falls der Spieler den Ball sieht. Sichergestellt, dass andere Spieler diese Information bekommen und gegebenenfalls verwenden.
 *
 * Revision 1.12  2002/01/22 18:04:00  jens
 * Some Staminaproblems solved
 *
 * Revision 1.11  2002/01/22 14:07:51  fastjack
 * Timing verbessert
 *
 * Revision 1.10  2002/01/22 11:01:55  fastjack
 * BTeamAction refactoring
 *
 * Revision 1.9  2002/01/18 19:49:28  seekamp
 * Prioritaeten werden durch Andys FileReader aus individuellen Spielertextdateien eingelesen.
 *
 * Revision 1.8  2002/01/03 01:04:36  fastjack
 * Parser verwendet fuer hear Messages wieder die originale CMU Methode. Ausserdem habe ich einige weitere Ausgaben in das PotentialField getan, um herauszufinden, wo er aussteigt. Dann habe ich das Timing so geHACKT, dass es immer true gibt, ausser es ist geblockt. Debugging necessary.
 *
 */
