/*****************************************************************************
 * $RCSfile: BallMessage.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/04/15 03:24:19 $
 * $Revision: 1.1 $
 *****************************************************************************/

#include "BallMessage.hh"
#include <cstdlib>
#include <iostream>

char BallMessage::itobase64char( int value ) {

    if ( value < 26 ) return (char)('a'+value);
    if ( value < 52 ) return (char)('A'+( value-26 ));
    if ( value < 62 ) return (char)('0'+( value-52 ));
    if ( value == 62 ) return '(';
    if ( value == 63 ) return ')';

    return '!';
}

void BallMessage::itobase64( int value, char *str ) {


    if (value > 63) {
        str[0] = itobase64char( value / 64 );
        str[1] = itobase64char( value % 64 );
        str[2] = 0;
    } else {
        str[0] = itobase64char( value );
        str[1] = 0;
    }

}

int BallMessage::base64chartoint( char ch ) {

    if ( ch == '(' ) return 62;
    if ( ch == ')' ) return 63;
    if ( ch < 'A' ) return (int)( ch - '0' )+52;
    if ( ch < 'a' ) return (int)( ch - 'A' )+26;

    return (int)( ch - 'a' );
}

int BallMessage::base64fromstr( char * str ) {

    int len = strlen( str );

    if (len == 1) {

        return (int)base64chartoint(str[0]);

    } else if ( len == 2 ) {

        int higher = base64chartoint(str[0]);
        int lower = base64chartoint(str[1]);

        return higher*64 + lower;

    } else return 0;

}

bool BallMessage::testConversionUtilities() {

    int z = 1;

    char str[10];
    itobase64( z,str );
    int erg = base64fromstr( str );

    if ( z != erg )
        std::cerr << "Test Output (BallMessage::testConversionUtilities)" << std::endl << "conversion " << z << " = " << erg << std::endl;

    return (z == erg);

}
BallMessage::BallMessage( char * str ) {

  if ( str[0] != IDENTIFIER )
     std::cerr << "BallMessage() trying to parse non-BUGS say message: " << str << std::endl;

  for ( int i=0; i<10; i++ ) this->str[i] = str[i];
  this->str[10] = 0;

  parseBallMessage( &str[1],x,y,conf,vx,vy,vconf,dist );

}
BallMessage::BallMessage( float x,
                          float y,
                          float conf,
                          float vx,
                          float vy,
                          float vconf,
                          float dist ) {

        this->x = x;
        this->y = y;
        this->conf = conf;
        this->vx = vx;
        this->vy = vy;
        this->vconf = vconf;
        this->dist = dist;

        int xUmgerechnet = (int)((x + 64.0)*32.0);
        int yUmgerechnet = (int)((y + 64.0)*32.0);
        if (dist > 63) dist = 63;
        if (dist < 0) dist = 0;
        int confUmgerechnet = (int)(conf*63);
        int vconfUmgerechnet = (int)(vconf*63);
        int xVel = (int)((vx + 3.0)*10.0);
        int yVel = (int)((vy + 3.0)*10.0);

        str[0] = IDENTIFIER;

        itobase64( xUmgerechnet, &str[1] );
        itobase64( yUmgerechnet, &str[3] );
        itobase64( (int)dist,         &str[5] );
        itobase64( confUmgerechnet,&str[6] );
        itobase64( xVel,         &str[7] );
        itobase64( yVel,         &str[8] );
        itobase64( vconfUmgerechnet,&str[9] );

}

void BallMessage::parseBallMessage( char * message,
                                    float & ballX,
                                    float & ballY,
                                    float & ballConf,
                                    float & ballVelX,
                                    float & ballVelY,
                                    float & ballVelConf,
                                    float & ballDist) {

    static char buf[4];

    buf[2] = 0;

    buf[0] = message[0];
    buf[1] = message[1];
    buf[2] = 0;
    ballX = ((float)base64fromstr( buf )/32.0)-64.0;

    buf[0] = message[2];
    buf[1] = message[3];
    buf[2] = 0;
    ballY = ((float)base64fromstr( buf )/32.0)-64.0;

    buf[0] = message[4];
    buf[1] = 0;
    ballDist = (float)base64fromstr( buf );

    buf[0] = message[5];
    buf[1] = 0;
    ballConf = (float)base64fromstr( buf )/63.0;

    buf[0] = message[6];
    buf[1] = 0;
    ballVelX = (float)base64fromstr( buf )/10.0 - 3.0;

    buf[0] = message[7];
    buf[1] = 0;
    ballVelY = (float)base64fromstr( buf )/10.0 - 3.0;

    buf[0] = message[8];
    buf[1] = 0;
    ballVelConf = (float)base64fromstr( buf )/63.0;
}

bool BallMessage::testParseBallMessage() {

    float ballX;
    float ballY;
    float ballConf;
    float ballVelX;
    float ballVelY;
    float ballVelConf;
    float ballDist;
    float testballX = 1;
    float testballY = 2;
    float testballConf = 0.95;
    float testballVelX = 2.3;
    float testballVelY = -1.8;
    float testballVelConf = 0.9;
    float testballDist = 23;

    BallMessage * m = new BallMessage( testballX,
                                       testballY,
                                       testballConf,
                                       testballVelX,
                                       testballVelY,
                                       testballVelConf,
                                       testballDist );

    parseBallMessage ( m->getCharString(),
                       ballX,
                       ballY,
                       ballConf,
                       ballVelX,
                       ballVelY,
                       ballVelConf,
                       ballDist );
    float eta = 0.2;

    bool res = ( ( testballX+eta > ballX ) && ( testballX-eta < ballX ) &&
                 ( testballY+eta > ballY ) && ( testballY-eta < ballY ) &&
                 ( testballConf+eta > ballConf ) && ( testballConf-eta < ballConf ) &&
                 ( testballVelX+eta > ballVelX ) && ( testballVelX-eta < ballVelX ) &&
                 ( testballVelY+eta > ballVelY ) && ( testballVelY-eta < ballVelY ) &&
                 ( testballVelConf+eta > ballVelConf ) && ( testballVelConf-eta < ballVelConf ) &&
                 ( testballDist+eta > ballDist ) && ( testballDist-eta < ballDist ) );

    if (!res) {

        std::cerr << "Test Output (BallMessage::testParseBallMessage)" << std::endl <<
            "Message " << m->getCharString() << std::endl <<
            "x " << testballX << "=" << ballX << std::endl <<
            "y " << testballY << "=" << ballY << std::endl <<
            "conf " << testballConf << "=" << ballConf << std::endl <<
            "vx " << testballVelX << "=" << ballVelX << std::endl <<
            "vy " << testballVelY << "=" << ballVelY << std::endl <<
            "vconf " << testballVelConf << "=" << ballVelConf << std::endl <<
            "dist " << testballDist << "=" << ballDist << std::endl;
    }

    return res;
}


/*
 * $Log: BallMessage.cc,v $
 * Revision 1.1  2002/04/15 03:24:19  fastjack
 * BallMessage Klasse fr das Parsen/Generieren von BallInfos gebaut. CommunicationManager aufgerumt.
 *
 */


