/*****************************************************************************
 * $RCSfile: Behaviour.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/04/09 19:23:14 $
 * $Revision: 1.6 $
 *****************************************************************************/

#include "Behaviour.hh"
#include <cmath>
#include "client.h"

cmu::GameTime Behaviour::opponentCloserToTheirGoalTime(0);
cmu::GameTime Behaviour::teammateCloserToTheirGoalTime(0);
bool Behaviour::opponentCloserToTheirGoal;
bool Behaviour::teammateCloserToTheirGoal;



ComparableRay Behaviour::leftTangentOfCircleThroughPoint(const cmu::Vector & center, float radius, const cmu::Vector & point) {

  cmu::Vector c = center;
  cmu::Vector p = point;

  float r2dFactor = ((float)180)/((float)M_PI);
  float dist = c.dist(p);
  
  cmu::AngleDeg globalAngle = ( c-p ).dir();
  cmu::AngleDeg relativeAngle = asin( radius / dist ) * r2dFactor;

  return ComparableRay(p, globalAngle-relativeAngle);
}

ComparableRay Behaviour::rightTangentOfCircleThroughPoint(const cmu::Vector & center, float radius, const cmu::Vector & point) {

  cmu::Vector c = center;
  cmu::Vector p = point;

  float r2dFactor = ((float)180)/((float)M_PI);
  float dist = c.dist(p);
  cmu::AngleDeg globalAngle = ( c-p ).dir();
  cmu::AngleDeg relativeAngle = asin( radius / dist ) * r2dFactor;

  return ComparableRay(p, globalAngle+relativeAngle);
}  

bool Behaviour::ballKickableForOpponent() {

  for (int i=1; i<=cmu::Mem->SP_team_size; i++){
    if ( cmu::Mem->OpponentPositionValid(i) &&
	 cmu::Mem->BallKickableForOpponent(i) )
      return true;
  }
  return false;
}

bool Behaviour::isPlayerCloserToTheirGoal() {

  if (isTeammateCloserToTheirGoal() ||
      isOpponentCloserToTheirGoal())   return true;

  else return false;

}

bool Behaviour::isOpponentCloserToTheirGoal() {
  static cmu::Vector theirGoalPosition( 52.0, 0.0 );

  if ( cmu::Mem->CurrentTime > opponentCloserToTheirGoalTime ) {
    
    float mydist = cmu::Mem->MyPos().dist( theirGoalPosition );

    opponentCloserToTheirGoal = false;
    for (int i=1; (i<=cmu::Mem->SP_team_size)&&(!opponentCloserToTheirGoal); i++) {

      if ( cmu::Mem->OpponentPositionValid(i) &&
	   cmu::Mem->OpponentAbsolutePosition(i).dist( theirGoalPosition ) < mydist )

	opponentCloserToTheirGoal = true;

    }

    opponentCloserToTheirGoalTime = cmu::Mem->CurrentTime;
    
  }
  
  return opponentCloserToTheirGoal;
}

bool Behaviour::isTeammateCloserToTheirGoal() {
  static cmu::Vector theirGoalPosition( 52.0, 0.0 );

  if ( cmu::Mem->CurrentTime > teammateCloserToTheirGoalTime ) {
    
    teammateCloserToTheirGoal = 
      (cmu::Mem->ClosestTeammateTo(theirGoalPosition) != cmu::Mem->MyNumber);

    teammateCloserToTheirGoalTime = cmu::Mem->CurrentTime;
    
  }
  

  return teammateCloserToTheirGoal;
}


/*
 * $Log: Behaviour.cc,v $
 * Revision 1.6  2002/04/09 19:23:14  fastjack
 * Verbesserungen der Dribble und Torschuss Methoden. Bug fixes.
 *
 * Revision 1.5  2002/04/08 17:59:47  fastjack
 * ScoringBehviour reimplemented. Does not work, yet. So don't worry if we don't shoot a goal...
 *
 * Revision 1.4  2002/04/05 16:15:42  fastjack
 * kleinere Optimierungen
 *
 * Revision 1.3  2002/03/25 23:40:00  fastjack
 * bug fix
 *
 * Revision 1.2  2002/03/25 18:17:03  fastjack
 * Attack hat noch BUGS :o)
 *
 * Revision 1.1  2002/03/25 16:14:06  fastjack
 * Merged & added Behaviour classes for offensive behaviour
 *
 */








