/*****************************************************************************
 * $RCSfile: Behaviour.hh,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *		       Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/04/09 19:23:14 $
 * $Revision: 1.9 $
 *
 * TODO: move verbessern
 *
 *****************************************************************************/

#ifndef BEHAVIOUR_H
#define BEHAVIOUR_H

#include "geometry.h"
#include "BTeamTypes.hh"
#include <list>

class ComparableRay : public cmu::Ray {

  static const float COORD_THRESHOLD = 0.5;
  static const float ANG_THRESHOLD = 2;

  bool isSimilar( cmu::Vector a, cmu::Vector b ) {
    return (((a.x + COORD_THRESHOLD) > (b.x)) &&
	    ((a.x - COORD_THRESHOLD) < (b.x)) &&
	    ((a.y - COORD_THRESHOLD) < (b.y)) &&
	    ((a.y + COORD_THRESHOLD) > (b.y)));
  }
  bool isSimilar( cmu::AngleDeg a, cmu::AngleDeg b ) {
    return (((a + ANG_THRESHOLD) > b) &&
	    ((a - ANG_THRESHOLD) < b));
  }

  public:
  ComparableRay() : cmu::Ray() { };
  ComparableRay(cmu::Vector orig, cmu::Vector dir) : cmu::Ray( orig, dir ) { };
  ComparableRay(cmu::Vector orig, float ang) : cmu::Ray( orig, ang ) { };

  bool operator ==(const ComparableRay &a) { return ((origin == a.origin) && (direction == a.direction)); };
  bool operator !=(const ComparableRay &a) { return !( (*this) == a ); };
  bool similarto(ComparableRay &a) { 
    return (isSimilar( origin, a.origin ) && isSimilar(direction.dir(), a.direction.dir())); 
  }
  
  void print() {
    cout << "(ComparableRay origin=" << origin << " direction=" << direction << ")" << endl;
  }

  inline cmu::Vector getOrigin() { return origin; }
  inline cmu::Vector getDirection() { return direction; }
  
  
};



class Behaviour {

  // cached values for performance reasons
  static cmu::GameTime opponentCloserToTheirGoalTime;
  static cmu::GameTime teammateCloserToTheirGoalTime;
  static bool opponentCloserToTheirGoal;
  static bool teammateCloserToTheirGoal;
   

  // Hilfsmethoden, die Praedikate und Werte fuer die Behaviours berechnen

  protected:
  ComparableRay leftTangentOfCircleThroughPoint(const cmu::Vector & center, float radius, const cmu::Vector & point);
  ComparableRay rightTangentOfCircleThroughPoint(const cmu::Vector & center, float radius, const cmu::Vector & point);

  bool ballKickableForOpponent();
  bool isPlayerCloserToTheirGoal();
  bool isOpponentCloserToTheirGoal();
  bool isTeammateCloserToTheirGoal();

  public:
  virtual bool behave() = 0; // abstract method
};


#endif

/*
 * $Log: Behaviour.hh,v $
 * Revision 1.9  2002/04/09 19:23:14  fastjack
 * Verbesserungen der Dribble und Torschuss Methoden. Bug fixes.
 *
 * Revision 1.8  2002/04/08 17:59:47  fastjack
 * ScoringBehviour reimplemented. Does not work, yet. So don't worry if we don't shoot a goal...
 *
 * Revision 1.7  2002/04/05 16:15:42  fastjack
 * kleinere Optimierungen
 *
 * Revision 1.6  2002/04/04 20:16:18  fastjack
 * moved PotentialField Behaviour in DefaultBehaviour class
 *
 * Revision 1.5  2002/03/26 17:32:11  seekamp
 * Es muessten jetzt Steilpaesse gespielt werden, konnte allerdings noch nicht getestet werden.
 *
 * Revision 1.4  2002/03/26 13:09:29  fastjack
 * Zugriffsmethoden hinzugefuegt
 *
 * Revision 1.3  2002/03/25 23:50:24  fastjack
 * bug fix
 *
 * Revision 1.2  2002/03/25 18:17:04  fastjack
 * Attack hat noch BUGS :o)
 *
 * Revision 1.1  2002/03/25 16:14:06  fastjack
 * Merged & added Behaviour classes for offensive behaviour
 *
 */
