/*****************************************************************************
 * $RCSfile: DefaultBehaviour.hh,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/04/08 13:32:09 $
 * $Revision: 1.8 $
 *****************************************************************************/

#ifndef _DEFAULTBEHAVIOUR_HH_
#define _DEFAULTBEHAVIOUR_HH_


#include "Behaviour.hh"

// forward declarations
namespace cmu {
  class Vector;
  class PlayerObject;
}
class Timing;

class DefaultBehaviour : public Behaviour {
  //

  bool checkAction(Action * action);

  /** Calculates the predicted velocity which the ball would get if we would kick it to
      the given target point. */
  cmu::Vector getPredictedBallVelocity(cmu::Vector & target);

  /** Bewertet einen Zielpunkt, ob ein Schuss dahin sicher wre. Dabei werden
          Kreise um den Gegner gezogen, und geprft ob die Linie zum Zielpunkt den Kreis
      schneidet. Diese Berechnung ist natrlich ungenau. Aber sie wird wesentlich
      schneller berechnet als eine vollstndige Interception Point Berechnung. */
  bool isSafeFromOpponent(cmu::Vector target, cmu::PlayerObject * opponent);


  /** Eine schnelle ungenaue Methode. Sie geht alle Gegner durch und schaut, ob
      einer in der Nhe des Weges zum Zielpunkt steht. Gibt nur true zurck, wenn
      alle einen gewissen Abstand zum Weg des Balles haben. Beachtet keine Mitspieler!
      */
  bool isSafeQuicklyChecked(Action * action);



  /** Bewertet eine Schuss-Aktion, ob sie sicher ist. Sicher heisst, dass der Ball
      nicht abgefangen werden kann. Dabei wird geprft, welcher Spieler am schnellsten
      am Ball sein kann. Dafr muss in das CMU Weltmodell eingegriffen werden, in der
      Art dass die Ballgeschwindigkeit und -richtung (also die Velocity), die vor dem
      Schuss natrlich noch 0 sind, auf einen geschtzten Wert gesetzt werden, bevor
      die Interception Point Berechnung von CMU luft.*/
  bool isSafe(Action * action);


  /** Zweck siehe isSafe. Implementation hnlich, nur anders. */
  bool isSafe2(Action * action);


  Action * selectPossibleAction(list<Action*> * actions);


  Action * getAction();


  PlayerInterceptionInfo calculateInterceptionInfoForPlayer( void * player,
                                                             Action * action );

  void forceAllIPCalculations();

  // =============== Attributes

  Timing * timing;
  bool debug;
  Action * currentAction;
  bool list_mode;

  public:
  DefaultBehaviour( Timing * t ) : timing( t ), debug( true ), currentAction( 0 ), list_mode( false ) {};

  virtual bool behave();

  void setDebug( bool b ) { debug = b; }

  bool isListMode() { return list_mode; }
  void setListMode( bool b ) { list_mode = b; }
  
};


#endif

/*
 * $Log: DefaultBehaviour.hh,v $
 * Revision 1.8  2002/04/08 13:32:09  fastjack
 * listMode jetzt als Property von DefaultBehaviour. per default auf false, Torwart setzt auf true.
 *
 * Revision 1.7  2002/04/04 21:15:43  fastjack
 * removed PotentialField code from BTeamAction, all is in DefaultBehaviour now. I hope it works.
 *
 * Revision 1.6  2002/04/04 20:44:04  fastjack
 * bug fix
 *
 * Revision 1.5  2002/04/04 20:42:14  fastjack
 * bug fix
 *
 * Revision 1.4  2002/04/04 20:37:44  fastjack
 * bug fix
 *
 * Revision 1.3  2002/04/04 20:23:25  fastjack
 * bug fix
 *
 * Revision 1.2  2002/04/04 20:16:18  fastjack
 * moved PotentialField Behaviour in DefaultBehaviour class
 *
 * Revision 1.1  2002/04/02 15:58:11  fastjack
 * Various changes. Torwart Abschlag Bugfix. kleinere unbedeutendere Korrekturen.
 *
 */
