/*****************************************************************************
 * $RCSfile: PFieldLogger.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: jens $
 * $Date: 2002/03/12 15:45:30 $
 * $Revision: 1.19 $
 *****************************************************************************/

#include "PFieldLogger.hh"
#include "Log.hh"
#include "world.hh"
#include "PotentialField.hh"
#include <string.h>
#include <stdio.h>
#include <fstream>

PFEntry::PFEntry() {};


PFieldLogger PFieldLogger::instance;

PFieldLogger::PFieldLogger() : enabled(true),initialized(true),count(0) {};

PFieldLogger & PFieldLogger::getInstance() {
  if (!instance.initialized) instance.init();

  return instance;
}

void PFieldLogger::init() {
  if (enabled) {

    fileCounter = 0;
    do {
       sprintf(filename,"/tmp/fields%d-%d.pf",cmu::Mem->MyNumber,fileCounter);
       ifstream ifs(filename);
       if (!ifs) {
	 break;
       } else {
	 ifs.close();
	 fileCounter++;
       }
    } while (true);

    initialized = true;
  }
}


void PFieldLogger::setEnabled(bool b) {
  enabled = b;
}

void PFieldLogger::log(PotentialField* potentialField) {
  if (enabled) {
    //Log::debug << "log (" << count << ")" << endl;
//    entries[count++]=potentialField->getPFEntry();

    entries[count++] = createEntryFromPotentialField(potentialField);

    //Log::debug << "log done (" << count << ")" << endl;
  }
}

void PFieldLogger::flush(bool b) {
  this->flush(b,0);
}

void PFieldLogger::flush(bool b, int unum) {
  if ((count >= 30) || b) { // nur dann was tun wenn auch Daten da sind...

    PFEntry* tmpEntry;
    if(unum>0)
    sprintf(filename,"/tmp/fields%d-%d.pf", unum,fileCounter);
    else
    sprintf(filename,"/tmp/fields%d-%d.pf",cmu::Mem->MyNumber,fileCounter);
    //Log::debug << "flushing to file " << filename << endl;
    of.open (filename);

    fileCounter++; // naechstes Mal andere Datei verwenden (mit hoeherem Zaehler)

    for (int i=0;i<count;i++) {

      tmpEntry = entries[i];
      of << (*tmpEntry) << '$';

    }

    of.close();

    // so alles geschrieben, buffer wieder von vorne fuellen.
    count = 0;

  }

}

ostream & operator<<(ostream & os, const ObjectEntry &oe) {
  // os << oe.unum << ',' << oe.x << ',' << oe.y << ',' << oe.dir;
  os << oe.x << ',' << oe.y << ',' << oe.dir<<',' <<oe.unum;
  return os;
}

ostream & operator<<(ostream & os, const PFEntry &pfe) {
  os << pfe.gametime << '|';
  os << pfe.playmode << '|';
  os << pfe.xRes << ',' << pfe.yRes << '|';
  for (int j=0;j<11;j++) {
    os << pfe.opponents[j];
    if (j<10) os << ' ';
  }
  os << "|";
  for (int j=0;j<10;j++) {
    os << pfe.teammates[j];
    if (j<9) os << ' ';
  }
  os << "|";
  os << pfe.ball << '|';
  os << pfe.self << '|';
  os << pfe.xTarget << ',' << pfe.yTarget << '|';
  os << pfe.tactic << '|';
  os << pfe.direction << '|';
  os << pfe.action << '|';
  os << pfe.stamina << '|';
  os << pfe.teampart << '|';
  os << pfe.situation << '|';
  os << pfe.closestPlayer << '|';
  for (int fieldy=0; fieldy<PotentialField::FIELD_SIZE_Y;fieldy++)
    for (int fieldx=0; fieldx<PotentialField::FIELD_SIZE_X;fieldx++)
      os << pfe.pfValues[fieldx][fieldy] << ' ';
  return os;
}


/**
* Diese Methode ist friend von PotentialField und darf somit auch auf die
* private Werte zugreifen.
*/
PFEntry * PFieldLogger::createEntryFromPotentialField(PotentialField * field) {

  PFEntry* temp = new PFEntry();
  World * world = field->world;

  temp->gametime=(world->currentTime.t+world->currentTime.s);
  temp->playmode=world->playMode;
  temp->xRes=PotentialField::FIELD_SIZE_X;
  temp->yRes=PotentialField::FIELD_SIZE_Y;
  int counter=0;
  for(int i = 0; i<world->getNumberOfTeammates();i++){
    if(world->team[i].isValid()){
      temp->teammates[counter].unum=world->team[i].unum;
      temp->teammates[counter].x   =world->team[i].x;
      temp->teammates[counter].y   =world->team[i].y;
      temp->teammates[counter].dir =(int)world->team[i].dir;
      //cout<<"x/y: ("<<world->team[i].x<<"/"<<world->team[i].y<<")"<<endl;
      
      counter++;
    }
  }
  //cout<<"Teammates geloggt"<<endl;
  counter = 0;
  for(int i = 0; i<world->getNumberOfOpponents();i++){
    if(world->opponent[i].isValid()){
      temp->opponents[counter].unum=world->opponent[i].unum;
      temp->opponents[counter].x   =world->opponent[i].x;
      temp->opponents[counter].y   =world->opponent[i].y;
      temp->opponents[counter].dir =(int)world->opponent[i].dir;
      counter++;
    }
  }
  //cout<<"Opponents geloggt"<<endl;
  temp->self.unum=world->myNumber;
  //cout<<"mynumber: "<<world->myNumber <<endl;
  temp->self.x=world->self.x;
  temp->self.y=world->self.y;
  temp->self.dir=(int)world->self.dir;
  temp->ball.unum=-2;
  temp->ball.x=world->ball.x;
  temp->ball.y=world->ball.y;
  temp->ball.dir=(int)world->ball.dir;
  temp->xTarget=field->xbest;
  temp->yTarget=field->ybest;
  temp->tactic=field->Taktik;
  //cout<<"before loc side"<<endl;
  temp->direction=world->myLocationSide;
  //cout<<"location side"<<endl;
  temp->action=field->Aktion[2];
  temp->stamina=world->self.stamina;
  temp->teampart=field->my_Mannschaftsteil;
  if(world->ballpossessing)temp->situation=1;
  else if(world->teamBallPossession)temp->situation=2;
  else if(world->closestPlayerToBall)temp->situation=3;
  else temp->situation=0;
  temp->closestPlayer=world->numOfClosestPlayerToBall;
  //cout<<"Feddich geloggt"<<endl;
  

  int nullcounter = 0;

  // FIXME
  for(int i=0;i<PotentialField::FIELD_SIZE_X;i++)
    for(int j =0;j<PotentialField::FIELD_SIZE_Y;j++) {
      if (field->fieldInfo[i][j] == 0) nullcounter++;
      temp->pfValues[i][j]=field->fieldInfo[i][j];
    }


  return temp;
}

/*
 * $Log: PFieldLogger.cc,v $
 * Revision 1.19  2002/03/12 15:45:30  jens
 * Grundlage fuer das Lernen der Prioritaeten implementiert. Fehler dabei ausgemerzt und Algorithmus fuer das Lernen auf Papier konkretesiert.
 *
 * Revision 1.18  2002/03/08 16:24:31  jens
 * Arbeiten beendet, die uns Potentialfeldszenen im Client parsen lassen und uns dann vollstaendige Potentialfelder genriert. Potentialfelder die generiert wurden und welche, die wir als Szene genommen haben (also die Feldwerte weglassen) stimmen nun ueberein. Rundungsfehler sind beseitigt. Somit sollte die Grundlage fuer die automatische Prioritaetenanpassung gelegt sein.
 *
 * Revision 1.17  2002/03/08 12:58:00  fastjack
 * code wieder bereinigt
 *
 * Revision 1.16  2002/03/05 17:08:01  fastjack
 * Interception Methoden fast fertig
 *
 * Revision 1.15  2002/03/01 11:18:46  fastjack
 * GPL added
 *
 * Revision 1.14  2002/02/26 17:59:30  jens
 * weiter am parsen eines eigenen PF gearbeitet. Einbinden in das Weltmodell funzt noch nicht richtig
 *
 * Revision 1.13  2002/02/11 17:04:15  seekamp
 * Fehlpassanalyse mit Wertanpassungen
 *
 * Revision 1.12  2002/02/10 12:19:13  fastjack
 * Habe die Includes angepasst, damit die Header sich so wenig wie mglich
 * gegenseitig includen. Z.T. ersetzt durch forward Deklaration. getPFEntry Methode
 * aus dem PotentialField heraus geholt und in den PFieldLogger gepackt - ist dort
 * sinniger.
 *
 * Revision 1.11  2002/02/05 18:19:39  jens
 * Fixed the Parse Sound. Now we can hear the Ball and the opponents broadcasted by other players.
 *
 * Revision 1.10  2002/01/29 19:34:00  fastjack
 * Konflikte behoben.
 *
 * Revision 1.9  2002/01/29 12:06:37  fastjack
 * debugging in progress
 *
 * Revision 1.8  2002/01/29 10:57:58  jens
 * kick verbessert und pausen der Spieler teilweise behoben (scan field with body eingesetzt)
 *
 * Revision 1.7  2002/01/25 17:22:35  fastjack
 * compiles under Linux AND Solaris8 now. so can be run on the sxx machines now.
 *
 * Revision 1.6  2002/01/21 12:08:36  fastjack
 * PFieldLogger: das Format so geaendert, dass der PFViewer die PFelder auch parsen kann. Deshalb die UNUM erstmal wieder entfernt
 *
 * Revision 1.5  2002/01/18 20:11:20  fastjack
 * Das Logging von Potentialfeldern fertig gestellt. Das Logging ist auch
 * ein-/ausschaltbar ( PFieldLogger::getInstance().setEnabled(true) ).
 *
 * Revision 1.4  2002/01/17 18:58:24  fastjack
 * flush() Methode optimiert
 *
 * Revision 1.3  2002/01/15 16:33:42  fastjack
 * Neue features fuer die Log-Klasse (kann jetzt wie cout etc verwendet werden), und kleinere Verbesserungen fuer den PFieldLogger hinzugefuegt.
 *
 * Revision 1.2  2002/01/15 10:32:48  fastjack
 * bug fixes
 *
 * Revision 1.1  2002/01/15 09:42:08  fastjack
 * Logging Klasse fuer das Abspeichern der Potentialfelder hinzugefuegt.
 *
 */
