/*****************************************************************************
 * $RCSfile: ScoringBehaviour.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 *
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: jens $
 * $Date: 2002/04/13 07:59:17 $
 * $Revision: 1.12 $
 *
 * TODO: move verbessern
 *
 *****************************************************************************/

#include "ScoringBehaviour.hh"
#include "geometry.h"
#include "mybehave.h"
#include "client.h"
#include "Log.hh"
#include "kick.h"


cmu::Vector ScoringBehaviour::goalie_pos()
{
  
  float goalie_x = 0.0;
  cmu::PlayerObject * tempPlayer;
  cmu::PlayerObject * goalie=0;
  
  
  for (int i=0; i<cmu::Mem->NumOpponents();i++) 
    {
      tempPlayer = cmu::Mem->Opponents()[i];
      
      if ((tempPlayer != NULL) &&
	  (tempPlayer->pos_valid())) {
	if (tempPlayer->get_abs_pos().x > cmu::Mem->BallX())
	  {
	    if (tempPlayer->get_abs_pos().x > goalie_x)
	      {
		goalie_x = tempPlayer->get_abs_pos().x;
		goalie = tempPlayer;
	      }
	  }
      }
    }  
  if (goalie_x==0)
    return cmu::Vector(36.0,0.0);
  
  return goalie->get_abs_pos();
}



bool ScoringBehaviour::isSituation() {
  /*
  - mypos.y weniger als +-20m von der mitte
  - mypos.x weniger als 20m von der grundlinie
    ODER
    mupos.x weniger als 10m vom torwart entfernt
   */

  return ( ( ( isNearToTheirGoal() ) || ( isNearToTheirGoalie() ) ) &&
	   ( isClosestPlayerToBall() ) && (cmu::Mem->GetBall()->kickable(0.0) ) );

}


bool ScoringBehaviour::isFlankenSituation()
{
  return ((cmu::Mem->MyX()>43 ) && ((cmu::Mem->MyY()>=20 ) || (cmu::Mem->MyY()<=-20))&&
	  ( isClosestPlayerToBall() ) && (cmu::Mem->GetBall()->kickable(0.0) ));
}







bool ScoringBehaviour::isClosestPlayerToBall() {
  
  return cmu::Mem->MyNumber==cmu::Mem->ClosestTeammateToBall(cmu::cmu_TRUE);
  
}


bool ScoringBehaviour::isNearToTheirGoalie() {
  
  //cmu::PlayerObject * goalie = cmu::Mem->GetPlayer( cmu::Mem->TheirSide, 1 );
  //if ( (goalie == 0) ) return false;
  cmu::Vector goalie = goalie_pos();
  Log::debug << "goalie x " << goalie.x << endl;
  return (( (goalie.x-cmu::Mem->MyX()) <= 8 ) &&( cmu::Mem->MyY() < 20 ) && ( cmu::Mem->MyY() > -20 ));
  
}

bool ScoringBehaviour::isNearToTheirGoal() {
  
  Log::debug << "isNearToTheirGoal " << cmu::Mem->MyX() << endl;
  return (( cmu::Mem->MyX() > 40 )&&( cmu::Mem->MyY() < 20 )&&( cmu::Mem->MyY() > -20 ));
  
}



bool ScoringBehaviour::isNearToBall() {

  cmu::BallObject * ball = cmu::Mem->GetBall();
  if ((ball == 0) || (!ball->pos_valid())) return false;

  return ( ( cmu::Mem->BallDistance() < 2 ) &&
	   ( cmu::Mem->MyNumber==cmu::Mem->ClosestTeammateToBall(cmu::cmu_TRUE)) );

}


float ScoringBehaviour::opponentGoalieInterceptionDistance(cmu::Vector goalCorner)
{
  cmu::Vector keeper = goalie_pos();
  float keeper_delta_x = keeper.x-cmu::Mem->MyX();
  float goal_delta_y = goalCorner.y-cmu::Mem->MyY();
  float goal_delta_x = goalCorner.x-cmu::Mem->MyX();

  if (goal_delta_x==0)
    {
      return 0.0;
    }
  float distance  = cmu::Mem->MyPos().dist(goalCorner);
  float interception = (keeper_delta_x/goal_delta_x)*goal_delta_y;
  if (interception<keeper.y)
    return (keeper.y-interception)/distance;
  else
    return (interception-keeper.y)/distance;
}






bool ScoringBehaviour::kickToGoalLeftCorner()
  {
    Log::debug << "Abstand Torwart Schnittpunkt links :: " << opponentGoalieInterceptionDistance(goalLeftCorner) << endl;
    Log::debug << "Abstand Torwart Schnittpunkt rechts :: " << opponentGoalieInterceptionDistance(goalRightCorner) << endl;
    
    return (opponentGoalieInterceptionDistance(goalLeftCorner)>opponentGoalieInterceptionDistance(goalRightCorner));
  }
 









/** Bewertet einen Zielpunkt, ob ein Schuss dahin sicher wre. Dabei werden
    Kreise um den Gegner gezogen, und geprft ob die Linie zum Zielpunkt den Kreis
    schneidet. Diese Berechnung ist natrlich ungenau. Aber sie wird wesentlich
    schneller berechnet als eine vollstndige Interception Point Berechnung. */
bool ScoringBehaviour::isSafeFromOpponent(cmu::Vector target, cmu::PlayerObject * opponent, float radiusFactor) {

  cmu::AngleDeg   angleToTarget     = (target - cmu::Mem->MyPos()).dir();
  cmu::Vector     opponentPosition  = opponent->get_abs_pos();
  float           distanceToTarget  = cmu::Mem->MyPos().dist(target);
  float           distanceToOpponent= cmu::Mem->MyPos().dist(opponentPosition);
  cmu::Vector     sol1,sol2;
  float           radius            = 2.0 * radiusFactor; // 0.75

  if (distanceToTarget < distanceToOpponent) return true;

  if (distanceToOpponent > 8.0) {
    radius = 4.0 * radiusFactor;

    return cmu::RayCircleIntersect( cmu::Ray( cmu::Mem->MyPos(), angleToTarget ),
                                    radius, opponentPosition,
                                    &sol1, &sol2 ) < 2;
  }

  return cmu::RayCircleIntersect( cmu::Ray( cmu::Mem->MyPos(), angleToTarget ),
                                  radius, opponentPosition,
                                  &sol1, &sol2 ) < 2;
}

void ScoringBehaviour::shoot( const cmu::Vector & target, bool hardest ) {

  Log::debug << "ScoringBehaviour::shoot()" << (hardest?" hard!":" carefully") << endl;

  if ( hardest ) {

    Action action( target );
    action.strength = 1;
    cmu::Mem->doKick( action );

  } else

    cmu::Mem->doKickToPoint( target, 2.7 );

}


/** This is the first method getting invoked when the player
    gets near enough to the goal. It has to evaluate the current
    situation: if it is possible to score, depending on the
    positions of the opponents. If it is indeed possible, it
    invokes shoot() directly. Otherwise it invokes avoidOpponent(),
    which has to check, in which direction we should kick the
    ball. After kicking the ball it puts this behaviour in the
    GETBALL state. The next time the getBall() method will run. 
void ScoringBehaviour::startScoring() {
  //
  Log::debug << "ScoringBehaviour::startScoring()" << endl;

  cmu::Vector target;

  if ( isPossibleToScore( target ) ) {
    Log::debug << "isPossibleToScore( " << target << " )" << endl;

    shoot( target );

  } else {
    Log::debug << "!isPossibleToScore( )" << endl;

    if (findTargetPoint( target )) {
      Log::debug << "target: " << target << " my pos: " << cmu::Mem->MyPos() << endl;

      shoot ( target, false );

      setScoringState( GETBALL );

      scoringInProgress = true;

    } else {
      // oops could not find good point, so just shoot...
      shoot( goalCenter );

    }

  }

}
*/

/**
    Calculates the point to which the player should kick if
    he wishes to avoid an opponent. We want to kick to the left if
    we are on the right side, and kick to the right if we are on
    the left side. But only if we can do it and no opponent is
    in the target area.
 */
bool ScoringBehaviour::findTargetPoint( cmu::Vector & target ) {
  float myy = cmu::Mem->MyY();
  float myx = cmu::Mem->MyX();
  float ty;
  float tx = myx - 2;
  cmu::Unum oppUnum;
  cmu::PlayerObject * opp;
  cmu::Unum secondOppUnum;
  cmu::PlayerObject * secondOpp;
  float firstDist;
  float secondDist;

  if (myy < 0)
    ty = myy + 15.0;
  else
    ty = myy - 15.0;
  cmu::Vector firstChoice( tx,ty );

  if (myy < 0)
    ty = myy - 15.0;
  else
    ty = myy + 15.0;
  cmu::Vector secondChoice( tx,ty );

  oppUnum = cmu::Mem->ClosestOpponentTo( firstChoice );
  if (oppUnum == Unum_Unknown) {
    target = firstChoice;
    return true;
  }

  opp = cmu::Mem->GetPlayer( cmu::Mem->TheirSide, oppUnum );
  if ( (opp == 0) || (!opp->pos_valid()) ) {
    target = firstChoice;
    return true;
  }

  secondOppUnum = cmu::Mem->ClosestOpponentTo( secondChoice );
  if ( secondOppUnum == Unum_Unknown ) secondOpp = opp;
  else
    secondOpp = cmu::Mem->GetPlayer( cmu::Mem->TheirSide, secondOppUnum );

  firstDist = opp->get_abs_pos().dist( firstChoice );
  secondDist = secondOpp->get_abs_pos().dist( secondChoice );

  if ( firstDist > 5.0  ) {
    target = firstChoice;
    return true;
  }
  if ( secondDist > 5.0 ) {
    target = secondChoice;
    return true;
  }
  if ( (firstDist-secondDist) < -2.0 ) {
    target = secondChoice;
    return true;
  }

  if ( firstDist < 3.0 ) {
    return false;
  }


  target = firstChoice;
  return true;


}




bool ScoringBehaviour::isPossibleToScore( cmu::Vector & target ) {

  static cmu::Vector firstChoice;
  static cmu::Vector secondChoice;


  // Wir pruefen zunaechst die Tormitte (sicherste Loesung, wenn moeglich), dann die kurze Ecke (kuerzester Weg) und dann
  // die lange Ecke


  if (cmu::Mem->MyY()>0)
    {
      firstChoice = goalRightCorner;
      secondChoice = goalLeftCorner;
    } else {
      firstChoice = goalRightCorner;
      secondChoice = goalLeftCorner;
    }


  int i = 0;
  float myx = cmu::Mem->MyX();

  bool centerSafe = true;
  bool firstSafe = true;
  bool secondSafe = true;

  while (i < cmu::Mem->NumOpponents() && (centerSafe || firstSafe || secondSafe)) {

    if (i < cmu::Mem->NumOpponents()) {
      cmu::PlayerObject * tempPlayer = cmu::Mem->Opponents()[i];

      if ((tempPlayer != 0) &&
	  (tempPlayer->pos_valid())) {
	
	if (!( tempPlayer->get_abs_pos().x < myx ) ) {
	
	  if (centerSafe && !isSafeFromOpponent( goalCenter, tempPlayer ) ) {
	    centerSafe = false;
	  }
	  if (firstSafe && !isSafeFromOpponent( firstChoice, tempPlayer ) ) {
	    firstSafe = false;
	  }
	  if (secondSafe && !isSafeFromOpponent( secondChoice, tempPlayer ) ) {
	    secondSafe = false;
	  }
	
	}
      }
    }

    i++;
  }

  if ( firstSafe ) {
    target = firstChoice;
  }
  else if ( secondSafe ) {
    target = secondChoice;
  }
  else if ( centerSafe ) {
    target = goalCenter;
  }

  if (centerSafe || firstSafe || secondSafe)
    Log::debug << "found target " << target << endl;
  else
    Log::debug << "no safe target found" << endl;

  return (centerSafe || firstSafe || secondSafe);
}

/*

void ScoringBehaviour::getBall() {

  Log::debug << "ScoringBehaviour::getBall()" << endl;

  if (cmu::Mem->BallKickable()) {
    scoringInProgress = false;

    // we have the ball and restart the scoring process
    setScoringState( START );
    startScoring();

  } else {
    scoringInProgress = true;
    cmu::get_ball();
  }

  }*/


/*

  fuere Behaviour aus, wenn
  - mypos.y weniger als +-20m von der mitte
  - mypos.x weniger als 20m von der grundlinie
    ODER
    mupos.x weniger als 5m vom torwart entfernt

  - Pruefe ob Torschuss moeglich
    - Wenn ja, direkt Schuss
    - Wenn nein,
      - finde ausweichpunkt
      - schiesse zum ausweichpunkt
      - getball

 */

bool ScoringBehaviour::behave() {
   

  if ( isSituation() ) { 

    if (!getball) {
      //
      
      if (!cmu::Mem->BallKickable()) {
	cmu::get_ball();
	return true;
      }
      else
	getball = true;
      
    }


    if (kickToGoalLeftCorner())
      {
	shoot( goalLeftCorner,false );
      } else {
	shoot( goalRightCorner,false );
      }



    return true;
  }

  if (isFlankenSituation())
    {
      
    if (!getball) {
      //
      
      if (!cmu::Mem->BallKickable()) {
	cmu::get_ball();
	return true;
      }
      else
	getball = true;
      
    }
        
    shoot(FlankenZielPunkt,false);
    return true;
    
    }
  
  
  

  return false;
}


bool ScoringBehaviour::testbehave() {

  static int nokickcounter = 0;
  static cmu::Vector nokicktarget;
  static int kickablecount = 0;
  
  if ( nokickcounter > 0 ) {
    nokickcounter--;
    cmu::go_to_point( nokicktarget );
    return true;
  }

  if ( scoringInProgress &&
       ( (!cmu::Mem->BallPositionValid()) ||
	 (cmu::Mem->BallDistance() > 15) ||
	 (ballKickableForOpponent()) ) ) {

    /*
      Falls nach dem Ausweichen der Ball verloren geht, soll der
      Ball-Verfolgungs-und-Schuss-Modus wieder aufgeloest werden
    */

    scoringInProgress = false;
    Log::debug << "no scoring" << endl;
  }
  


  if ( scoringInProgress ) {
    Log::debug << "ball verfolgung" << endl;
    
    /*
      Nach dem Ausweichen soll dieser Fall ausgefuehrt werden:
      - falls nicht kickable, get_ball
      - ansonsten schuss nach:
      - bester zielpunkt oder pauschal mittelpunkt
     */
    
    if ( cmu::Mem->BallKickable() ) {

      
      cmu::Vector target;
      
      if ( isPossibleToScore( target ) ) {
	
	shoot( target );
	
      } else {

	if (cmu::Mem->MyY()>0)
	  {
	    shoot( goalRightCorner );
	  } else {
	    shoot( goalLeftCorner );
	  }

      }

      scoringInProgress = false;

    } else

      cmu::get_ball();
    
    return true;
  }


  if ( isSituation() ) {
    Log::debug << "scoring!" << (cmu::Mem->BallKickable()?" kickable":" NOT HAVE BALL") << endl;
   /*
      - Pruefe ob Torschuss moeglich
       - Wenn ja, direkt Schuss
       - Wenn nein,
        - finde ausweichpunkt
        - schiesse zum ausweichpunkt

     */
    
    cmu::Vector target;
    
    if ( isPossibleToScore( target ) ) {
      Log::debug << "check ok, shoot to goal" << endl;
      
      shoot( target );
      
    } else {
      Log::debug << "cannot score, weiching aus" << endl;
      
      if ( kickablecount == 1 ) {
	cmu::stop_ball();
	kickablecount++;
	return true;
      }
      if ( kickablecount == 2 ) {
	cmu::get_ball();
	kickablecount++;
	return true;
      }
      if ( kickablecount == 0 ) {
	cmu::get_ball();
	kickablecount++;
	return true;
      }
      else kickablecount = 0;
      
      if (findTargetPoint( target )) {

	kick_ball( target, cmu::KM_HardestKick );

	// shoot( target, false );
	scoringInProgress = true;
	nokickcounter = 4;
	nokicktarget = target;
	
	Log::debug << "ausgeweicht, verfolging ball" << endl;
	
      } else {
	
	Log::debug << "cannot ausweich" << endl;
	if (cmu::Mem->MyY()>0)
	  {
	    shoot( goalRightCorner );
	  } else {
	    shoot( goalLeftCorner );
	  }

      }
      

    }

    return true;
    
  }

  return false;
}


/*
 * $Log: ScoringBehaviour.cc,v $
 * Revision 1.12  2002/04/13 07:59:17  jens
 * German Open 2002 Edition
 *
 * Revision 1.11  2002/04/11 00:00:42  fastjack
 * Warnungen beseitigt
 *
 * Revision 1.10  2002/04/10 13:24:53  seekamp
 * Dribbling/Schuss verbessert.
 * Steilpaesse werden weniger hart gespielt.
 * Wenn aussen wird Richtung, Eckfahne gedribbelt
 *
 * Revision 1.9  2002/04/09 19:23:14  fastjack
 * Verbesserungen der Dribble und Torschuss Methoden. Bug fixes.
 *
 * Revision 1.8  2002/04/09 17:07:55  seekamp
 * kleine Korrekturen
 *
 * Revision 1.7  2002/04/09 00:27:45  fastjack
 * kick improved
 *
 * Revision 1.6  2002/04/08 17:59:47  fastjack
 * ScoringBehviour reimplemented. Does not work, yet. So don't worry if we don't shoot a goal...
 *
 * Revision 1.5  2002/04/08 14:40:51  fastjack
 * Scoring Absturz Bugfix
 *
 * Revision 1.4  2002/04/05 14:56:55  seekamp
 * In AttackBehaviour das Laufen zum Zielpunkt eines Steilpasses zielstrebiger gestaltet.
 *
 * Revision 1.3  2002/04/04 15:54:07  seekamp
 * kleinere Veraenderungen am Attack und Shot Behaviour zudem Libero wieder eingefuehrt und Aussenverteidiger bei gegnerischen Angriffen ueber die gegenueberliegende Seite Richtung Mitte gezogen.
 *
 * Revision 1.2  2002/04/02 17:06:51  fastjack
 * Scoring Behaviour ungetesterweise eingebaut.
 *
 * Revision 1.1  2002/04/02 15:58:11  fastjack
 * Various changes. Torwart Abschlag Bugfix. kleinere unbedeutendere Korrekturen.
 *
 */
