/*****************************************************************************
 * $RCSfile: Timing.hh,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/03/05 17:08:01 $
 * $Revision: 1.14 $
 *****************************************************************************/

#ifndef _TIMING_HH_
#define _TIMING_HH_

class Timing {

  private:
  static const int MAX_IDLE_CYCLES = 30;
  static const float NEAR_TO_BALL = 10.0;
  static const float MID_TO_BALL = 48.0;

  /** noch zu blockierende zyklen */
  int blockedCycles;
  /** zaehlt die kicks waehrend der blockierung */
  int kicksWhileBlockedState;
  
  bool seen;
  bool didGetBall;
  int nextEvalTime;  
  
  /**
   * Falls das Erzeugen eines Potentialfeldes aus irgendwelchen
   * technischen Gruenden gerade verboten ist, gibt diese Methode
   * true zurueck.
   */
  bool blocked();

  /**
   * Die Funktion ermittelt, ob sehr dringend eine Bewertung 
   * durchgefuehrt werden muss. Das ist der Fall, wenn:
   * - wir nah am Ball sind
   * - vielleicht noch mehr
   */
  bool isHighPriority();

  /**
   * Die Funktion ermittelt, ob wir weniger dringend
   * bewerten muessen.
   */
  bool isMidPriority();

  /**
   * Die Funktion ermittelt, ob wir eher spaeter die Bewertung
   * durchfuehren koennen.
   */
  bool isLowPriority();
  

  public:
  Timing() : blockedCycles(0), kicksWhileBlockedState(0), seen(false), didGetBall(false), nextEvalTime(0) {}
  
  void printDebugInfo();
  
  /**
   * Berechnet, ob der Agent in der derzeitigen Situation 
   * sein Potentialfeld neu erzeugen darf/muss oder nicht.
   * @return true, falls ja, false sonst
   */
  bool allowToGenerate();

  void registerGeneration();
  inline void setNextEvalTime(int gt) { nextEvalTime = gt; }

  void setSeen(bool b);
  bool getSeen();

  /** has to be called from low level action component
      sets counter to a certain number of cycles in which the
      agent should NOT try to run after the ball
  */
  void registerKick() {
    if (blockedCycles > 0) kicksWhileBlockedState++;
    else {  
      blockedCycles = 4;
      kicksWhileBlockedState = 0;
    }
  }
  /** has to be called from low level action component
      sets a flag which overrides the timing and runs the
      generation
  */
  void registerGetBall() { didGetBall = true; }
  
  /** decides if a player is allowed to kick in this situation
      gives true if not blocked
  */
  bool mayKick() { return kicksWhileBlockedState < 3; }
  
};


#endif

/*
 * $Log: Timing.hh,v $
 * Revision 1.14  2002/03/05 17:08:01  fastjack
 * Interception Methoden fast fertig
 *
 * Revision 1.13  2002/03/01 11:18:47  fastjack
 * GPL added
 *
 * Revision 1.12  2002/02/07 20:07:51  fastjack
 * merged
 *
 * Revision 1.11  2002/01/31 18:22:26  fastjack
 * Versuch einer getBall Optimierung
 *
 * Revision 1.10  2002/01/29 13:55:50  fastjack
 * debugging still in progress
 *
 * Revision 1.9  2002/01/29 11:13:35  fastjack
 * Timing update
 *
 * Revision 1.8  2002/01/25 18:48:32  seekamp
 * Staminaverwaltung implementiert
 *
 * Revision 1.7  2002/01/22 14:07:51  fastjack
 * Timing verbessert
 *
 * Revision 1.6  2002/01/22 12:06:13  fastjack
 * Timing tests
 *
 * Revision 1.5  2002/01/17 18:15:30  seekamp
 * Verschiedene Methoden in Potentialfield und BTeamAction veraendert, mit dem
 * Ergebnis, dass der ballnaechste Spieler zum Ball laeuft und einen Pass spielt.
 * (mehr oder weniger)
 *
 * Revision 1.4  2001/12/20 18:16:24  jens
 * eleminated different compile errors and started to implement simple actions
 *
 * Revision 1.3  2001/12/18 18:43:11  fastjack
 * Compiled nicht. Vorsicht, habe eine Menge auskommentiert. Timing z.B., und alles wo cmu:: drin vorkommt. Kruzifix!
 *
 * Revision 1.2  2001/12/14 20:41:40  fastjack
 * PotentialField neu geschrieben auf Basis von Matthias' Code
 *
 * Revision 1.1  2001/12/13 20:39:01  fastjack
 * Timing Klasse entwickelt. Ist noch Alpha :-)
 *
 */
