/*****************************************************************************
 * $RCSfile: bugsobjects.hh,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: jens $
 * $Date: 2002/03/08 16:24:31 $
 * $Revision: 1.12 $
 *****************************************************************************/

#ifndef BUGSOBJECTS_HH
#define BUGSOBJECTS_HH

#include "utils.h"
#include "geometry.h"
#include "cmutypes.h"
#include "Log.hh"

struct Point  {
  int x;
  int y;
};

class AgingObject {
  static const int MAX_AGE = 1;
  
  /** gets calculated while update */
  int age;
  
  protected:
  /** cycle of last see */
  cmu::GameTime seenLast;
  /** calculates values, if no current values are known */
  virtual void estimate() = 0; // abstract method
  
  public:
  inline int getAge() { return age; };
  /** returns false, if the seen data is too old to make sense */
  inline bool isValid() { return (age < MAX_AGE); }
  /** returns a factor which says, how sure the value is
      the older the seen data, the lower the value
      if isValid gives false, then sureness is 0
  */
  float sureness();
  
  /** gets called every time the field should be generated
      it checks, how old the data is and calls estimate() if necessary */
  void update(cmu::GameTime current);
};


class BasicObject : public AgingObject {
  private:
  /** position of last see */
  cmu::Vector seenPos;
  /* velocity of last see (not needed now) */
  cmu::Vector seenVel;
  
  /** [calculates x and y values, if no current values are known] 
      umm... doesn't do anything useful at the moment because CMU
      already estimates.
  */
  virtual void estimate();
  /** bekommt die x und y Koordinaten des Spielers aus dem Weltmodell
      von CMU und rechnet diese um in die Koordinaten des
      PotentialFields */
  void setXY(float x, float y);
  
  public:
  
  static const float CONFIDENCY = 0.7;
  
  int x;
  int y;
  float dir;
  int altesx;
  int altesy;
  cmu::Unum unum;
  
  /** collects seen data. has to be called before update. */
  void set(cmu::GameTime current, 
		  cmu::Vector pos, 
		  cmu::Unum _unum 
		  /*, cmu::Vector & vel=0*/ );
  
  void set(cmu::GameTime current, 
		  cmu::Vector pos, 
		  float _dir, 
		  cmu::Unum _unum );
  void set(cmu::GameTime current, int newX, int newY, float _dir, cmu::Unum _unum ) ;
  

  /** gibt den Wert zurueck, wie er aus dem CMU Modell gekommen war */
  inline cmu::Vector & getPosition() { return seenPos; }
};

class SelfObject : public BasicObject {
  public:
  int stamina;
};
class BallObject : public BasicObject {
  public:
  void set(cmu::GameTime current, 
	   cmu::Vector pos, 
	   float _dir);
  void set(cmu::GameTime current,int newX, int newY, float _dir );
};


  
#endif

/*
 * $Log: bugsobjects.hh,v $
 * Revision 1.12  2002/03/08 16:24:31  jens
 * Arbeiten beendet, die uns Potentialfeldszenen im Client parsen lassen und uns dann vollstaendige Potentialfelder genriert. Potentialfelder die generiert wurden und welche, die wir als Szene genommen haben (also die Feldwerte weglassen) stimmen nun ueberein. Rundungsfehler sind beseitigt. Somit sollte die Grundlage fuer die automatische Prioritaetenanpassung gelegt sein.
 *
 * Revision 1.11  2002/03/08 15:23:10  fastjack
 * Neue .set Methode eingefuehrt zum uebergeben von am Raster orientierten Koordinaten
 *
 * Revision 1.10  2002/03/01 11:18:48  fastjack
 * GPL added
 *
 * Revision 1.9  2002/02/25 12:15:32  fastjack
 * Abfangen von Durchgebrochenen Gegnern eingebaut.
 *
 * Revision 1.8  2002/02/05 18:19:40  jens
 * Fixed the Parse Sound. Now we can hear the Ball and the opponents broadcasted by other players.
 *
 * Revision 1.7  2002/02/05 13:41:08  fastjack
 * gemerged
 *
 * Revision 1.6  2002/02/05 12:27:41  jens
 * Weltmodell update
 *
 * Revision 1.5  2002/01/31 18:22:27  fastjack
 * Versuch einer getBall Optimierung
 *
 * Revision 1.4  2002/01/29 19:34:00  fastjack
 * Konflikte behoben.
 *
 * Revision 1.3  2002/01/28 11:42:04  fastjack
 * optimierungen
 *
 * Revision 1.2  2002/01/28 09:44:44  fastjack
 * link errors fixed
 *
 * Revision 1.1  2002/01/28 02:13:50  fastjack
 * Added an intermediary World object which should guess positions when it is done. Not ready yet, WARNING - code may or may not work.
 *
 */
