/*****************************************************************************
 * $RCSfile: world.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/04/01 15:57:13 $
 * $Revision: 1.24 $
 *****************************************************************************/

#include "world.hh"
#include "PotentialField.hh"
#include "Log.hh"
#include "client.h"

void World::hearTarget(float x, float y, int gt) {
  targetX = x;
  targetY = y;
  targetGT = gt;
}

void World::update(bool b) {
  // diese Methode hiess frher readWorldModelData()
  
  if(b){
    this->b = b;
    
    //nun muessen wir wissen, welche Scene aus dem vector wir nutzen wollen.
    //dazu muessen wir zusaetzlich eine Variable uebergeben, die die
    //zu nutzende Scene angibt. Da der vector bereits existiert, muessen wir 
    //lediglich die Variable bei jedem Aufruf uebergeben. Dazu gibt es die 
    //Methode "identScene(int index)". Sie wird aus client1.C aufgerufen
    //um die Scene anzugeben. Im folgenden werden dann die Werte ausgelesen 
    //und die Bewertung anzustossen.
    importObjectsFromScene();
    // cout<<"importobjects fin"<<endl;
    importOwnValuesFromScene();
    // cout<<"importown fin"<<endl;
    calculateAdditionalValues();
    // cout<<"calc add fin"<<endl;
    calculateOffsideLines();
    cmu::Mem->setInitialValues(myNumber);
    // cout<<"calc offside fin"<<endl;
  } else { 
    importFromCmu();
    calculateAdditionalValues();
    calculateOffsideLines();
  } 
}

void World::loadScene(vector<PField> pfield){
  scenes=pfield;
  //   cout<<"TM world.cc (" << scenes[1].tm[0].dir<< ")"<<endl;
  // cout<<"TM2 world.cc (" << pfield[1].tm[0].dir<< ")"<<endl;
}

void World::identScene(int index){
  identedScene = index;
}


void World::importFromCmu()
{
  importObjectsFromCmu();
  importOwnValuesFromCmu();
}

float World::translateToGlobalX(int x) {
  return  (ORIGINAL_FIELD_SIZE_X*2/(FIELD_SIZE_X))*x-ORIGINAL_FIELD_SIZE_X;
}

float World::translateToGlobalY(int y) {
  return  (ORIGINAL_FIELD_SIZE_Y*2/(FIELD_SIZE_Y))*y-ORIGINAL_FIELD_SIZE_Y;
}

//Diese Methode importiert eine Scene aus dem Vector. Sie schreibt die 
//Werte daraus in die Variablen des Weltmodels, etc...
void World::importObjectsFromScene(){
  cmu::GameTime currentTime = scenes[identedScene].zyklus;
  numberOfTeammates = 0;
  numberOfOpponents = 0;
  // cout<<"IdentedScene: ("<<identedScene<<") k:"<<scenes[identedScene].k <<endl;
   for (int i=0; i<scenes[identedScene].k; i++) {
     
    // ********* Teammates ********
      team[numberOfTeammates++].set( currentTime,
				     (scenes[identedScene].tm[i].x), 
				     (scenes[identedScene].tm[i].y),
				     scenes[identedScene].tm[i].dir,
				     scenes[identedScene].tm[i].unum);
   }
   // ********* Opponents ********
   for (int i=0; i<scenes[identedScene].j; i++) {
     opponent[numberOfOpponents++].set( currentTime,
					(scenes[identedScene].opp[i].x), 
					(scenes[identedScene].opp[i].y),
					scenes[identedScene].opp[i].dir,
					scenes[identedScene].opp[i].unum);
   }
    // Log::debug<<"Number of opponents in world.cc: ("<<numberOfOpponents <<")"<<endl;
 // Log::debug<<"Number of teammates in world.cc: ("<<numberOfTeammates <<")"<<endl;
   for (int i=0;i<MAX_NUMBER_OF_TEAMMATES;i++) {
     team[i].update(currentTime);
   }
   for (int i=0;i<11;i++) {
     opponent[i].update(currentTime);
   }
   if (scenes[identedScene].ball.x>-1)
     {
      ball.set( currentTime,
		(scenes[identedScene].ball.x),
		(scenes[identedScene].ball.y),
		scenes[identedScene].ball.dir);
    }
  ball.update(currentTime);
  
}


void World::importObjectsFromCmu() {
   // diese Methode hiess frher readPlayerInfoFromModelData()
  cmu::GameTime currentTime = cmu::Mem->CurrentTime;

  cmu::PlayerObject * tempPlayer = NULL;

  numberOfTeammates = 0;
  numberOfOpponents = 0;


  for (int i=0; i<MAX_PLAYERS; i++) {

    // ********* Teammates ********

    tempPlayer = cmu::Mem->Teammates()[i];

    if ((i < cmu::Mem->NumTeammates()) &&
	(tempPlayer != NULL) &&
	(tempPlayer->pos_valid()) &&
	(cmu::Mem->MyNumber != tempPlayer->unum)) {

      team[numberOfTeammates++].set( currentTime,
				     tempPlayer->get_abs_pos(),
				     tempPlayer->get_abs_body_ang(),
				     tempPlayer->unum );
    }

    // ********* Opponents ********

    tempPlayer = cmu::Mem->Opponents()[i];

    if ((i < cmu::Mem->NumOpponents()) &&
	(tempPlayer != NULL) &&
	(tempPlayer->pos_valid())) {

      opponent[numberOfOpponents++].set( currentTime,
					 tempPlayer->get_abs_pos(),
					 tempPlayer->get_abs_body_ang(),
					 tempPlayer->unum );
    }

    // ********* Teamless ********

    tempPlayer = cmu::Mem->TeamlessPlayers()[i];

    if ((i < cmu::Mem->NumTeamlessPlayers()) && (tempPlayer != NULL) &&
	(tempPlayer->pos_valid())) {

      opponent[numberOfOpponents++].set( currentTime,
					 tempPlayer->get_abs_pos(),
					 tempPlayer->get_abs_body_ang(),
					 tempPlayer->unum );
    }
  }

  //  Log :: debug << "number of teammates: " << numberOfTeammates << endl;
  //  Log :: debug << "number of opponents: " << numberOfOpponents << endl;
  //  Log :: debug << "number of teamless players: " << cmu::Mem->NumTeamlessPlayers() << endl;

  for (int i=0;i<MAX_NUMBER_OF_TEAMMATES;i++) {
    team[i].update(currentTime);
  }

  for (int i=0;i<MAX_NUMBER_OF_OPPONENTS;i++) {
    opponent[i].update(currentTime);
  }
  // Log::debug<<"Number of opponents in world.cc: ("<<numberOfOpponents <<")"<<endl;
  if (cmu::Mem->BallPositionValid())
    {

      //      cmu::Vector ballPosition = cmu::Mem->GetBall()->estimate_future_pos(int(cmu::Mem->BallDistance() / PREDICT_BALL_POS_STEPS));
      ball.set( currentTime,
		cmu::Mem->GetBall()->get_abs_pos(),
		cmu::Mem->GetBall()->get_abs_heading() );

    }
  ball.update(currentTime);


}

void World::importOwnValuesFromScene(){
  currentTime = scenes[identedScene].zyklus;

  ballpossessing     = (scenes[identedScene].spielsituation == 1);
  numOfClosestPlayerToBall = scenes[identedScene].closest;
  //cout<<"numOfClosestPlayerToBall: ("<<numOfClosestPlayerToBall<<")"<<endl;
  playMode           = (cmu::PM_Play_On);
  if(scenes[identedScene].spielsituation==2)teamBallPossession=true;
  if(scenes[identedScene].spielsituation==3)closestPlayerToBall=true;
  //closestPlayerToBall= (scenes[identedScene].closest);
  //ballPossessor      = (closestPlayerToBall);
  midOfOwnGoal       = (cmu::Vector(-52.0,0.0));
  //myPosition         = (cmu::Vector(translateToGlobalX(scenes[identedScene].ownPos.x),translateToGlobalY(scenes[identedScene].ownPos.y)));
  // TODO: check if CMU guesses own position if not seen. do we have to guess ?
  self.set(currentTime,(scenes[identedScene].ownPos.x),
	   (scenes[identedScene].ownPos.y),
	   scenes[identedScene].ownPos.dir,
	   scenes[identedScene].ownPos.unum);
  myNumber = scenes[identedScene].ownPos.unum;
  myLocationSide = scenes[identedScene].direction;
  self.stamina = scenes[identedScene].stamina;
  self.update(currentTime);
}


void World::importOwnValuesFromCmu() {

  teamBallPossession        = (cmu::Mem->BallPossessor() > 0);
  ballpossessing            = (cmu::Mem->MyNumber == cmu::Mem->BallPossessor())&&(cmu::Mem->BallKickable());

  playMode                  = (cmu::Mem->PlayMode);

  closestPlayerToBall       = (cmu::Mem->MyNumber==cmu::Mem->ClosestTeammateToBall(cmu::cmu_TRUE));

  cmu::Unum bpUnum          = (cmu::Mem->BallPossessor());
  if (bpUnum != Unum_Unknown)
    ballPossessor           = (cmu::Mem->GetOpponent( bpUnum ));
  else
    ballPossessor           = NULL;

  midOfOwnGoal              = (cmu::Vector(-52.0,0.0));
  myPosition                = (cmu::Mem->MyPos());

  //  if (cmu::Mem->MyNumber==1)
  // Log::info << "closest Player to ball? " << (closestPlayerToBall?"yes":"no") <<  endl;

  // TODO: check if CMU guesses own position if not seen. do we have to guess ?
  self.set( cmu::Mem->CurrentTime,
	    cmu::Mem->MyPos(),
	    cmu::Mem->MyBodyAng(),
	    cmu::Mem->MyNumber );
  self.stamina = (int)cmu::Mem->MyStamina();
  //  myNumber = cmu::Mem->MyNumber; // wird in BTeamAction::initialize gesetzt
  self.update(cmu::Mem->CurrentTime);
  myLocationSide = cmu::Mem->MyLocationSide();

  currentTime = cmu::Mem->CurrentTime;

  teammateCloserToTheirGoal = (cmu::Mem->ClosestTeammateTo(theirGoalPosition) != myNumber);

}

void World::calculateOffsideLines() {
  //
  int first, second, tmp;
  
  first = second = PotentialField::FIELD_SIZE_X/2;
  
  for ( int i=0; i < numberOfTeammates; i++) {
    if ( team[i].isValid() && (team[i].x < second) ){
      second = team[i].x;
      if (second < first) {
	tmp = first;
	first = second;
        second = tmp;
      }
    }
  }

  // jetzt ist first das kleinste x, second das zweitkleinste

  if ( ball.x < second )
    theirOffsideLine = ball.x;
  else
    theirOffsideLine = second;
  
  first = second = PotentialField::FIELD_SIZE_X/2;
  
  for ( int i=0; i < numberOfOpponents; i++) {
    if ( opponent[i].isValid() && (opponent[i].x > second) ){
      second = opponent[i].x;
      if (second > first) {
	tmp = first;
	first = second;
        second = tmp;
      }
    }
  }

  // jetzt ist first das kleinste x, second das zweitkleinste

  if ( ball.x > second )
    myOffsideLine = ball.x;
  else
    myOffsideLine = second;  
}

bool World::isMyOffsidePosition(int x) {
  // falls in eigener Haelfte: kein Abseits
  if (x < PotentialField::FIELD_SIZE_X/2) return false;

  return x > myOffsideLine;
}
bool World::isTheirOffsidePosition(int x) {
  // falls in deren Haelfte: kein Abseits
  if (x > PotentialField::FIELD_SIZE_X/2) return false;

  return x < theirOffsideLine;
}


World::World() : teammateCloserToTheirGoal(false), theirGoalPosition(52.5,0), myNumber( 0 ), myOffsideLine(PotentialField::FIELD_SIZE_X), theirOffsideLine(PotentialField::FIELD_SIZE_X) {};


void World::calculateAdditionalValues() {
  teamBallInMannschaftsteil = 2;
  oppBallInMannschaftsteil  = 2;

  if ((ball.x>0) & !(teamBallPossession))
    oppBallInMannschaftsteil  = 3;

  if ((ball.x>(int)(0.33*PotentialField::FIELD_SIZE_X)) & !(teamBallPossession))
    oppBallInMannschaftsteil  = 2;

  if ((ball.x>(int)(0.66*PotentialField::FIELD_SIZE_X)) & !(teamBallPossession))
    oppBallInMannschaftsteil  = 1;
  // cout<<"before po"<<endl;
  if(!b){ 
    cmu::PlayerObject * po = cmu::Mem->getClosestPlayerObjectTo ( cmu::Mem->BallAbsolutePosition() );
    numOfClosestPlayerToBall = (po==NULL)?0:po->unum;
    
  }
  //else numOfClosestPlayerToBall = closestPlayerToBall;
  // cout<<"after po"<<endl;
}

bool World::amIClosestPlayerToBall() {
  cmu::PlayerObject * tempPlayerObject;
  cmu::Vector ballPos = cmu::Mem->BallAbsolutePosition();
  tempPlayerObject = cmu::Mem->ClosestTeammateObjectTo(ballPos);
  if (tempPlayerObject != NULL) {
    float my_dist = ballPos.dist2(cmu::Mem->MyPos());
    float other_dist = ballPos.dist2(tempPlayerObject->get_abs_pos());
    return my_dist <= other_dist;
  } else {

    // Falls kein Mitspieler gesehen wurde, und ich selbst nah genug am Ball
    // dran bin, geben wir hier true zurueck, damit der Spieler dann zum
    // Ball laeuft.
    // Was haben wir hier fuer Werte ?? 300 = GERATEN :-) FIXME

    if (ballPos.dist2(cmu::Mem->MyPos()) <= 300.0||closestPlayerToBall==myNumber)
      return true;
    else
      return false;
  }
}

bool World::isPlayerCloserToTheirGoal() {
  if (isTeammateCloserToTheirGoal()) return true;

  for(int i=1;i<getNumberOfOpponents();i++)
  {
    if ((opponent[i].isValid()) & (opponent[i].x>self.x))
    return true;
  }

  return false;
}

bool World::isTeammateCloserToTheirGoal() {
  return teammateCloserToTheirGoal;
}

bool World::isMyStandardSituation() {
  // cout<<"isMyStandardSituation()"<<endl;
  
  switch (playMode)
    {
    case cmu::PM_My_Kick_Off:
      return true;
    case cmu::PM_My_Kick_In:
      return true;
    case cmu::PM_My_Free_Kick:
      return true;
    case cmu::PM_My_Corner_Kick:
      return true;
    case cmu::PM_My_Goal_Kick:
      return true;
    case cmu::PM_My_Goalie_Free_Kick:
      return true;
    case cmu::PM_My_Offside_Kick:
      return true;
    default:
      return false;
    }
  return false;
}
bool World::isTheirStandardSituation() {
  // cout<<"isTheirStandardSituation()"<<endl;
  switch (playMode)
    {
    case cmu::PM_Their_Kick_Off:
      return true;
    case cmu::PM_Their_Kick_In:
      return true;
    case cmu::PM_Their_Free_Kick:
      return true;
    case cmu::PM_Their_Corner_Kick:
      return true;
    case cmu::PM_Their_Goal_Kick:
      return true;
    case cmu::PM_Their_Goalie_Free_Kick:
      return true;
    case cmu::PM_Their_Offside_Kick:
      return true;
    default:
      return false;
    }
  return false;
}

bool World::isDefendingGoal() {
  if (ball.isValid()) {

    if (myPosition.x < (ball.getPosition().x - 1)) return true;
    if (ball.x <= 1) return true;

    // if (ball.x > (PotentialField::FIELD_SIZE_X/4)) return true;
    //    else return false;
    cmu::Vector d = ((ball.getPosition()-midOfOwnGoal).Normalize() - (myPosition-midOfOwnGoal).Normalize());
    cmu::AngleDeg dirDiff = d.dir();
    // Log::debug << "dirDiff = " << dirDiff << endl;
    return (dirDiff < 10)&&(dirDiff > -10);
  } else
    return true; // Wenn wir keinen Ball sehen, tun wir einfach so als
  // wuerden wir erfolgreich das tor verteidigen.
} 

bool World::isNoChanceToGetBall() {
  /* Hier wollen wir erkennen, ob ein Gegner auf der Aussenseite
     "durch" ist, und keinen Verteidiger mehr vor sich hat.
     Eigenschaften der Situation:
     - kein Mitspieler zwischen Gegner und der Linie da wo das Tor drauf is
     - Gegner ist im aeusseren Bereich
     - Gegner hat den Ball
     - Gegner bewegt sich in Richtung Linie-wo-das-Tor-drauf-is
  */

  //  if (cmu::Mem->MyNumber == 11) return true;

  // der aeussere Bereich:
  int grenzY =  22;
  int grenzX =  -22;

  /*  BasicObject * pfBallPossessor = NULL;

  if (ballPossessor == NULL) // mh keiner hat den Ball grad? 
    return false; 

  // Mitspieler hat Ball, also haben wir sehr wohl ne Chance.
  if (ballPossessor->side == cmu::Mem->MySide) return false; 

  // Falls sich der Gegner nicht in diese Richtung bewegt, 
  if ((ballPossessor->get_abs_heading() < -85)|| 
      (ballPossessor->get_abs_heading() > 85)) 
    return false; 

  if (ballPossessor->unum == 0) return false; 
  */
  
  cmu::Unum fastestTeammate = cmu::Mem->FastestTeammateToBall();
  cmu::Unum fastestOpponent = cmu::Mem->FastestOpponentToBall();

  if (fastestOpponent == Unum_Unknown) return false;

  cmu::AngleDeg dir = cmu::Mem->OpponentAbsoluteBodyAngle(fastestOpponent);
  // Falls sich der Gegner nicht in diese Richtung bewegt, 
  if ((dir < -85) || (dir > 85)) 
    return false; 
  
  float opponentX = cmu::Mem->OpponentAbsolutePosition(fastestOpponent).x;
  float opponentY = cmu::Mem->OpponentAbsolutePosition(fastestOpponent).y;

  // Falls sich Gegner ausserhalb des Bereiches befindet, gilt der Fall nicht
  if ((opponentX > grenzX) || 
      ((opponentY < (34-grenzY)) &&
       (opponentY > (-34+grenzY)))) return false;

  int teammateCycles = (fastestTeammate == Unum_Unknown?1000:
			cmu::Mem->TeammateInterceptionNumberCycles(fastestTeammate));
  int myCycles = cmu::Mem->MyInterceptionNumberCycles();
  int opponentCycles = cmu::Mem->OpponentInterceptionNumberCycles(fastestOpponent);

  if ((myCycles <= teammateCycles) && (myCycles > opponentCycles)) return true;
  return false; 
}

/*
 * $Log: world.cc,v $
 * Revision 1.24  2002/04/01 15:57:13  fastjack
 * world.update wird nicht mehr im PotentialField aufgerufen. Grund: mehr Flexibilitaet. AttackBehaviour und BTeamAction von World entkoppelt.
 *
 * Revision 1.23  2002/03/26 17:56:23  fastjack
 * Performance Optimierungen.
 *
 * Revision 1.22  2002/03/21 18:25:51  fastjack
 * Wieder die Auswahl von mehreren Punkten eingefuegt. Methode zur optimalen Ermittlung der Ball Velocity entwickelt.
 *
 * Revision 1.21  2002/03/15 13:26:01  fastjack
 * Ausgaben entfernt / dem Gouraud Kerl mehr Genauigkeit verschafft und einen boesen BUG entfernt
 *
 * Revision 1.20  2002/03/12 15:45:30  jens
 * Grundlage fuer das Lernen der Prioritaeten implementiert. Fehler dabei ausgemerzt und Algorithmus fuer das Lernen auf Papier konkretesiert.
 *
 * Revision 1.19  2002/03/11 11:55:56  fastjack
 * bugfix: Spieler haben falsche Player??.txt eingelesen
 *
 * Revision 1.18  2002/03/08 16:24:31  jens
 * Arbeiten beendet, die uns Potentialfeldszenen im Client parsen lassen und uns dann vollstaendige Potentialfelder genriert. Potentialfelder die generiert wurden und welche, die wir als Szene genommen haben (also die Feldwerte weglassen) stimmen nun ueberein. Rundungsfehler sind beseitigt. Somit sollte die Grundlage fuer die automatische Prioritaetenanpassung gelegt sein.
 *
 * Revision 1.17  2002/03/01 11:18:48  fastjack
 * GPL added
 *
 * Revision 1.16  2002/02/28 11:27:27  fastjack
 * Ausgaben entfernt
 *
 * Revision 1.15  2002/02/26 17:59:30  jens
 * weiter am parsen eines eigenen PF gearbeitet. Einbinden in das Weltmodell funzt noch nicht richtig
 *
 * Revision 1.14  2002/02/25 12:15:32  fastjack
 * Abfangen von Durchgebrochenen Gegnern eingebaut.
 *
 * Revision 1.13  2002/02/15 17:45:35  fastjack
 * PotentialField vorbereitet fuer Automatisierung durch zukuenftige Lernmethode, verschiedene Dinge aus World statt aus Mem geholt etc.
 *
 * Revision 1.12  2002/02/10 12:19:14  fastjack
 * Habe die Includes angepasst, damit die Header sich so wenig wie mglich
 * gegenseitig includen. Z.T. ersetzt durch forward Deklaration. getPFEntry Methode
 * aus dem PotentialField heraus geholt und in den PFieldLogger gepackt - ist dort
 * sinniger.
 *
 * Revision 1.11  2002/02/07 19:41:34  jens
 * Kommunikation auf Mitspieler und im Falle eines Schusses auch den Zielpunkt erweitert.
 *
 * Revision 1.10  2002/02/07 15:16:10  fastjack
 * mini update
 *
 * Revision 1.9  2002/02/05 13:41:08  fastjack
 * gemerged
 *
 * Revision 1.8  2002/02/05 12:27:41  jens
 * Weltmodell update
 *
 * Revision 1.7  2002/01/31 18:22:27  fastjack
 * Versuch einer getBall Optimierung
 *
 * Revision 1.6  2002/01/29 19:34:00  fastjack
 * Konflikte behoben.
 *
 * Revision 1.5  2002/01/29 19:18:51  seekamp
 * Verhalten fuer das Ausfuehren von Standardsituationen implementiert
 *
 * Revision 1.4  2002/01/28 17:22:03  seekamp
 * Dem Torwart Abschlaege und Abstoesse beigebracht
 *
 * Revision 1.3  2002/01/28 11:42:04  fastjack
 * optimierungen
 *
 * Revision 1.2  2002/01/28 09:44:44  fastjack
 * link errors fixed
 *
 * Revision 1.1  2002/01/28 02:13:50  fastjack
 * Added an intermediary World object which should guess positions when it is done. Not ready yet, WARNING - code may or may not work.
 *
 */
