/*****************************************************************************
 * $RCSfile: world.hh,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/04/01 15:57:13 $
 * $Revision: 1.16 $
 *****************************************************************************/

#ifndef WORLD_HH
#define WORLD_HH

#include "bugsobjects.hh"
#include "geometry.h"
//#include <vector.h>
#include "Generating.hh"

/**
This class manages all information from CMUs world model.
It especially remembers data from the past and guesses values on
this basis.
*/
class World {

  /** man gehe nicht auf die jetzige Position des Balles zu, 
      sondern auf die Pos die der Ball in einigen Zyklen haben wrde. */
  static const int PREDICT_BALL_POS_STEPS = 4;
  static const int MAX_NUMBER_OF_TEAMMATES = 10;
  static const int MAX_NUMBER_OF_OPPONENTS = 21;
  static const int FIELD_SIZE_X = 90;
  static const int FIELD_SIZE_Y = 60;
  static const float ORIGINAL_FIELD_SIZE_X = 52.5;
  static const float ORIGINAL_FIELD_SIZE_Y = 34;
  int numberOfTeammates;
  int numberOfOpponents;
  bool teammateCloserToTheirGoal;

  cmu::Vector theirGoalPosition;

  public:
  World();

  BasicObject team[MAX_NUMBER_OF_TEAMMATES];
  BasicObject opponent[MAX_NUMBER_OF_OPPONENTS];

  vector <PField> scenes;
  int teamBallInMannschaftsteil;
  int oppBallInMannschaftsteil;
  bool teamBallPossession;

  /** is true if we have the ball */
  bool ballpossessing;

  /** is true if we are the closes player to the ball (of our team
      ... we don't check the opponents) */
  bool closestPlayerToBall;
  int numOfClosestPlayerToBall;
  int myNumber;
  int myLocationSide;
  bool b;

  int myOffsideLine;
  int theirOffsideLine;
  int identedScene;

  BallObject ball;
  SelfObject self;

  float targetX;
  float targetY;
  cmu::GameTime targetGT;
  cmu::GameTime currentTime;
  cmu::Pmode playMode;

  // ******* Die folgenden Variablen sind im realen Koordinatensystem *******
  cmu::PlayerObject * ballPossessor;
  cmu::Vector midOfOwnGoal;
  cmu::Vector myPosition;

  // ******* Methoden

  inline int getNumberOfOpponents() { return numberOfOpponents; }
  inline int getNumberOfTeammates() { return numberOfTeammates; }

  bool isMyStandardSituation();
  bool isTheirStandardSituation();
  bool amIClosestPlayerToBall();
  bool isTeammateCloserToTheirGoal();
  bool isPlayerCloserToTheirGoal();
  bool isMyOffsidePosition(int x);
  bool isTheirOffsidePosition(int x);
  void identScene(int index);
  /** Falls Ball Possessor unbekannt ist, gibt diese Methode den Vector
      (FIELD_SIZE_X/2,FIELD_SIZE_Y/2) zurueck */
  inline cmu::Vector getBallPossessorPosition() {
    if ( ballPossessor == NULL ) return cmu::Vector(0,0);
    return ballPossessor->get_abs_pos();
  }

  /** Besondere Situation: Hier wollen wir erkennen, ob ein Gegner 
      auf der Aussenseite "durch" ist, und keinen Verteidiger 
      mehr vor sich hat. 
  */
  bool isNoChanceToGetBall();

  /** Situation: Wir stehen auf der Linie zwischen Ball und Tor */
  bool isDefendingGoal();

  inline bool isStandardSituation() { return isMyStandardSituation() || isTheirStandardSituation(); }

  void update(bool b);
  inline void importFromCmu();
  void importObjectsFromCmu();
  float translateToGlobalX(int x);
  float translateToGlobalY(int y);
  void importOwnValuesFromCmu();
  void loadScene(vector<PField> pfield);
  void importObjectsFromScene();
  void importOwnValuesFromScene();
  void calculateAdditionalValues();
  void calculateOffsideLines();

  void hearTarget(float x, float y, int gt);
};

#endif

/*
 * $Log: world.hh,v $
 * Revision 1.16  2002/04/01 15:57:13  fastjack
 * world.update wird nicht mehr im PotentialField aufgerufen. Grund: mehr Flexibilitaet. AttackBehaviour und BTeamAction von World entkoppelt.
 *
 * Revision 1.15  2002/03/26 17:56:23  fastjack
 * Performance Optimierungen.
 *
 * Revision 1.14  2002/03/15 13:26:01  fastjack
 * Ausgaben entfernt / dem Gouraud Kerl mehr Genauigkeit verschafft und einen boesen BUG entfernt
 *
 * Revision 1.13  2002/03/01 11:18:48  fastjack
 * GPL added
 *
 * Revision 1.12  2002/02/28 11:27:28  fastjack
 * Ausgaben entfernt
 *
 * Revision 1.11  2002/02/26 17:59:30  jens
 * weiter am parsen eines eigenen PF gearbeitet. Einbinden in das Weltmodell funzt noch nicht richtig
 *
 * Revision 1.10  2002/02/25 12:15:32  fastjack
 * Abfangen von Durchgebrochenen Gegnern eingebaut.
 *
 * Revision 1.9  2002/02/15 17:45:35  fastjack
 * PotentialField vorbereitet fuer Automatisierung durch zukuenftige Lernmethode, verschiedene Dinge aus World statt aus Mem geholt etc.
 *
 * Revision 1.8  2002/02/10 12:19:14  fastjack
 * Habe die Includes angepasst, damit die Header sich so wenig wie mglich
 * gegenseitig includen. Z.T. ersetzt durch forward Deklaration. getPFEntry Methode
 * aus dem PotentialField heraus geholt und in den PFieldLogger gepackt - ist dort
 * sinniger.
 *
 * Revision 1.7  2002/02/07 19:41:34  jens
 * Kommunikation auf Mitspieler und im Falle eines Schusses auch den Zielpunkt erweitert.
 *
 * Revision 1.6  2002/02/07 15:16:11  fastjack
 * mini update
 *
 * Revision 1.5  2002/02/05 13:41:08  fastjack
 * gemerged
 *
 * Revision 1.4  2002/02/05 12:27:41  jens
 * Weltmodell update
 *
 * Revision 1.3  2002/01/29 19:34:01  fastjack
 * Konflikte behoben.
 *
 * Revision 1.2  2002/01/28 09:44:45  fastjack
 * link errors fixed
 *
 * Revision 1.1  2002/01/28 02:13:50  fastjack
 * Added an intermediary World object which should guess positions when it is done. Not ready yet, WARNING - code may or may not work.
 *
 */
